#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Martin Hawlisch, Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2013       Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Import from GeneWeb"

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------
import re
import time

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".ImportGeneWeb")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.utils.libformatting import ImportInfo

_ = glocale.translation.gettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.errors import GedcomError, GrampsImportError
from gramps.gen.lib import (
    Attribute,
    AttributeType,
    ChildRef,
    Citation,
    Date,
    DateError,
    Event,
    EventRef,
    EventRoleType,
    EventType,
    Family,
    FamilyRelType,
    Name,
    NameType,
    Note,
    Person,
    PersonRef,
    Place,
    PlaceName,
    Source,
    LdsOrd,
)
from gramps.gen.db import DbTxn
from html.entities import name2codepoint

_date_parse = re.compile(
    r"([kmes~?<>]+)?([0-9/]+)([J|H|F])?(\.\.)?([0-9/]+)?([J|H|F])?"
)
_text_parse = re.compile(r"0\((.*)\)")

_mod_map = {
    ">": Date.MOD_AFTER,
    "<": Date.MOD_BEFORE,
    "~": Date.MOD_ABOUT,
}

_cal_map = {
    "J": Date.CAL_JULIAN,
    "H": Date.CAL_HEBREW,
    "F": Date.CAL_FRENCH,
}

pevents_map = {
    "#birt": EventType.BIRTH,  # Epers_Birth
    "#bapt": EventType.BAPTISM,  # Epers_Baptism
    "#deat": EventType.DEATH,  # Epers_Death
    "#buri": EventType.BURIAL,  # Epers_Burial
    "#crem": EventType.CREMATION,
    "#acco": EventType((EventType.CUSTOM, _("Accomplishment"))),
    "#acqu": EventType((EventType.CUSTOM, _("Acquisition"))),
    "#adhe": EventType((EventType.CUSTOM, _("Adhesion"))),
    "#awar": EventType((EventType.CUSTOM, _("Award"))),
    "#bapl": LdsOrd.BAPTISM,  # Epers_BaptismLDS
    "#barm": EventType.BAR_MITZVAH,  # Epers_BarMitzvah
    "#basm": EventType.BAS_MITZVAH,  # Epers_BatMitzvah
    "#bles": EventType.BLESS,  # Epers_Benediction
    "#cens": EventType.CENSUS,
    "#chgn": EventType((EventType.CUSTOM, _("Change Name"))),
    "#circ": EventType((EventType.CUSTOM, _("Circumcision"))),
    "#conf": EventType.CONFIRMATION,  # Epers_Confirmation
    "#conl": LdsOrd.CONFIRMATION,  # Epers_ConfirmationLDS
    "#degr": EventType.DEGREE,
    "#demm": EventType((EventType.CUSTOM, _("Military Demobilisation"))),
    "#dist": EventType((EventType.CUSTOM, _("Award"))),
    "#dotl": LdsOrd.ENDOWMENT,  # Epers_DotationLDS
    "#educ": EventType.EDUCATION,  # Epers_Education
    "#elec": EventType.ELECTED,  # Epers_Election
    "#emig": EventType.EMIGRATION,
    "#endl": EventType((EventType.CUSTOM, _("Dotation"))),
    "#exco": EventType((EventType.CUSTOM, _("Excommunication"))),
    "#fcom": EventType.FIRST_COMMUN,
    "#flkl": EventType((EventType.CUSTOM, _("LDS Family Link"))),
    "#fune": EventType((EventType.CUSTOM, _("Funeral"))),
    "#grad": EventType.GRADUATION,
    "#hosp": EventType((EventType.CUSTOM, _("Hospitalisation"))),
    "#illn": EventType((EventType.CUSTOM, _("Illness"))),
    "#immi": EventType.IMMIGRATION,
    "#lpas": EventType((EventType.CUSTOM, _("List Passenger"))),
    "#mdis": EventType((EventType.CUSTOM, _("Military Distinction"))),
    "#mobm": EventType((EventType.CUSTOM, _("Militaty Mobilisation"))),
    "#mpro": EventType((EventType.CUSTOM, _("Military Promotion"))),
    "#mser": EventType.MILITARY_SERV,  # Epers_MilitaryService
    "#natu": EventType.NATURALIZATION,  # Epers_Naturalisation
    "#occu": EventType.OCCUPATION,  # Epers_Occupation
    "#ordn": EventType.ORDINATION,  # Epers_Ordination
    "#prop": EventType.PROPERTY,  # Epers_Property
    "#resi": EventType.RESIDENCE,  # Epers_Residence
    "#reti": EventType.RETIREMENT,
    "#slgc": EventType(
        (EventType.CUSTOM, _("LDS Seal to child"))
    ),  # Epers_ScellentChildLDS
    "#slgp": LdsOrd.SEAL_TO_PARENTS,  # Epers_ScellentParentLDS
    "#slgs": LdsOrd.SEAL_TO_SPOUSE,
    "#vteb": EventType((EventType.CUSTOM, _("Sold property"))),  # Epers_VenteBien
    "#will": EventType.WILL,  # Epers_Will
}

fevents_map = {
    "#marr": EventType.MARRIAGE,  # Efam_Marriage
    "#nmar": EventType.NUM_MARRIAGES,
    "#nmen": EventType((EventType.CUSTOM, _("No mention"))),  # Efam_NoMention
    "#enga": EventType.ENGAGEMENT,  # Efam_Engage
    "#div": EventType.DIVORCE,
    "#sep": EventType((EventType.CUSTOM, _("Separated"))),  # Efam_Separated
    "#anul": EventType.ANNULMENT,  # Efam_Annulation
    "#marb": EventType.MARR_BANNS,  # Efam_MarriageBann
    "#marc": EventType.MARR_CONTR,  # Efam_MarriageContract)
    "#marl": EventType.MARR_LIC,  # Efam_MarriageLicense)
    "#resi": EventType.RESIDENCE,  # Efam_Residence)
}


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def importData(database, filename, user):
    global callback

    try:
        g = GeneWebParser(database, filename)
    except IOError as msg:
        user.notify_error(_("%s could not be opened\n") % filename, str(msg))
        return

    try:
        status = g.parse_geneweb_file()
    except IOError as msg:
        errmsg = _("%s could not be opened\n") % filename
        user.notify_error(errmsg, str(msg))
        return
    return ImportInfo({_("Results"): _("done")})


# -------------------------------------------------------------------------
# For a description of the file format see
# http://cristal.inria.fr/~ddr/GeneWeb/en/gwformat.htm
# https://github.com/geneanet/geneweb/issues/315
# -------------------------------------------------------------------------
class GeneWebParser:
    def __init__(self, dbase, file):
        self.db = dbase
        if file:  # Unit tests can create the parser w/o underlying file
            self.f = open(file, "rb")
            self.filename = file
            self.encoding = "iso-8859-1"
            self.gwplus = False

    def get_next_line(self):
        self.lineno += 1
        line = self.f.readline()

        try:
            line = line.decode("utf-8")
        except GrampsImportError as err:
            self.errmsg(str(err))

        if line:
            try:
                line = str(line.strip())
            except UnicodeDecodeError:
                line = line.decode(self.encoding).strip()
        else:
            line = None
        return line

    def parse_geneweb_file(self):
        with DbTxn(_("GeneWeb import"), self.db, batch=True) as self.trans:
            self.db.disable_signals()
            t = time.time()
            self.lineno = 0
            self.index = 0
            self.fam_count = 0
            self.indi_count = 0

            self.fkeys = []
            self.ikeys = {}
            self.pkeys = {}
            self.skeys = {}

            self.current_mode = None
            self.current_family = None
            self.current_husband_handle = None
            self.current_child_birthplace_handle = None
            self.current_child_source_handle = None
            try:
                while 1:
                    line = self.get_next_line()
                    if line is None:
                        break
                    if line == "":
                        continue

                    fields = line.split(" ")

                    LOG.debug("LINE: %s" % line)

                    if fields[0] == "gwplus":
                        self.gwplus = True
                        self.encoding = "utf-8"
                    elif fields[0] == "encoding:":
                        self.encoding = fields[1]
                    elif fields[0] == "fam":
                        self.current_mode = "fam"
                        self.read_family_line(line, fields)
                    elif fields[0] == "rel":
                        self.current_mode = "rel"
                        self.read_relationship_person(line, fields)
                    elif fields[0] == "src":
                        self.read_source_line(line, fields)
                    elif fields[0] in ("wit", "wit:"):
                        self.read_witness_line(line, fields)
                    elif fields[0] == "cbp":
                        self.read_children_birthplace_line(line, fields)
                    elif fields[0] == "csrc":
                        self.read_children_source_line(line, fields)
                    elif fields[0] == "beg" and self.current_mode == "fam":
                        self.read_children_lines()
                    elif fields[0] == "beg" and self.current_mode == "rel":
                        self.read_relation_lines()
                    elif fields[0] == "comm":
                        self.read_family_comment(line, fields)
                    elif fields[0] == "notes":
                        self.read_person_notes_lines(line, fields)
                    elif fields[0] == "fevt" and self.current_mode == "fam":
                        # self.read_fevent_line(self.get_next_line())
                        pass
                    elif fields[0] == "pevt":
                        # self.read_pevent_line(self.get_next_line(), fields)
                        pass
                    elif fields[0] == "notes-db":
                        self.read_database_notes_lines(line, fields)
                    elif fields[0] == "pages-ext" or "wizard-note":
                        pass
                    elif fields[0] == "end":
                        self.current_mode = None
                    else:
                        LOG.warning(
                            "parse_geneweb_file(): Token >%s< unknown. line %d skipped: %s"
                            % (fields[0], self.lineno, line)
                        )
            except GedcomError as err:
                self.errmsg(str(err))

            t = time.time() - t
            # Translators: leave all/any {...} untranslated
            msg = ngettext(
                "Import Complete: {number_of} second",
                "Import Complete: {number_of} seconds",
                t,
            ).format(number_of=t)

        self.db.enable_signals()
        self.db.request_rebuild()

        LOG.debug(msg)
        LOG.debug("Families: %d" % len(self.fkeys))
        LOG.debug("Individuals: %d" % len(self.ikeys))
        return None

    def read_family_line(self, line, fields):
        self.current_husband_handle = None
        self.current_child_birthplace_handle = None
        self.current_child_source_handle = None
        self.current_family = Family()
        self.db.add_family(self.current_family, self.trans)
        # self.db.commit_family(self.current_family,self.trans)
        self.fkeys.append(self.current_family.get_handle())
        idx = 1

        LOG.debug("\nHusband:")
        (idx, husband) = self.parse_person(fields, idx, Person.MALE, None)
        if husband:
            self.current_husband_handle = husband.get_handle()
            self.current_family.set_father_handle(husband.get_handle())
            self.db.commit_family(self.current_family, self.trans)
            husband.add_family_handle(self.current_family.get_handle())
            self.db.commit_person(husband, self.trans)
        LOG.debug("Marriage:")
        idx = self.parse_marriage(fields, idx)
        LOG.debug("Wife:")
        (idx, wife) = self.parse_person(fields, idx, Person.FEMALE, None)
        if wife:
            self.current_family.set_mother_handle(wife.get_handle())
            self.db.commit_family(self.current_family, self.trans)
            wife.add_family_handle(self.current_family.get_handle())
            self.db.commit_person(wife, self.trans)
        return None

    def read_relationship_person(self, line, fields):
        LOG.debug(r"\Relationships:")
        (idx, person) = self.parse_person(fields, 1, Person.UNKNOWN, None)
        if person:
            self.current_relationship_person_handle = person.get_handle()

    def read_relation_lines(self):
        if not self.current_relationship_person_handle:
            LOG.warning("Unknown person for relationship in line %d!" % self.lineno)
            return None
        rel_person = self.db.get_person_from_handle(
            self.current_relationship_person_handle
        )
        while 1:
            line = self.get_next_line()
            if line is None or line == "end":
                break
            if line == "":
                continue

            # match relationship type and related person
            line_re = re.compile("^- ([^:]+): (.*)$")
            matches = line_re.match(line)
            if matches:
                # split related person into fields
                fields = matches.groups()[1].split(" ")
                if fields:
                    (idx, asso_p) = self.parse_person(fields, 0, Person.UNKNOWN, None)
                    pref = PersonRef()
                    pref.set_relation(matches.groups()[0])
                    LOG.warning("TODO: Handle association types properly")
                    pref.set_reference_handle(asso_p.get_handle())
                    rel_person.add_person_ref(pref)
                    self.db.commit_person(rel_person, self.trans)
                else:
                    LOG.warning("Invalid name of person in line %d" % self.lineno)
            else:
                LOG.warning("Invalid relationship in line %d" % self.lineno)
                break
        self.current_mode = None
        return None

    def read_source_line(self, line, fields):
        if not self.current_family:
            LOG.warning("Unknown family of child in line %d!" % self.lineno)
            return None
        source = self.get_or_create_source(self.decode(fields[1]))
        self.current_family.add_citation(source.get_handle())
        self.db.commit_family(self.current_family, self.trans)
        return None

    def read_witness_line(self, line, fields):
        LOG.debug("Witness:")
        if fields[1] == "m:":
            (idx, wit_p) = self.parse_person(fields, 2, Person.MALE, None)
        elif fields[1] == "f:":
            (idx, wit_p) = self.parse_person(fields, 2, Person.FEMALE, None)
        else:
            (idx, wit_p) = self.parse_person(fields, 1, None, None)
        if wit_p:
            mev = None
            # search marriage event
            for evr in self.current_family.get_event_ref_list():
                ev = self.db.get_event_from_handle(evr.get_reference_handle())
                if ev.get_type() == EventType.MARRIAGE:
                    mev = ev  # found.
            if not mev:  # No marriage event found create a new one
                mev = self.create_event(EventType.MARRIAGE, None, None, None, None)
                mar_ref = EventRef()
                mar_ref.set_reference_handle(mev.get_handle())
                self.current_family.add_event_ref(mar_ref)
            wit_ref = EventRef()
            wit_ref.set_role(EventRoleType(EventRoleType.WITNESS))
            wit_ref.set_reference_handle(mev.get_handle())
            wit_p.add_event_ref(wit_ref)
            self.db.commit_person(wit_p, self.trans)
        return None

    def read_children_lines(self):
        father_surname = "Dummy"
        if not self.current_husband_handle:
            LOG.warning("Unknown father for child in line %d!" % self.lineno)
            return None
        husb = self.db.get_person_from_handle(self.current_husband_handle)
        father_surname = husb.get_primary_name().get_surname()
        if not self.current_family:
            LOG.warning("Unknown family of child in line %d!" % self.lineno)
            return None
        while 1:
            line = self.get_next_line()
            if line is None:
                break
            if line == "":
                continue

            fields = line.split(" ")
            if fields[0] == "-":
                LOG.debug("Child:")
                child = None
                if fields[1] == "h":
                    (idx, child) = self.parse_person(
                        fields, 2, Person.MALE, father_surname
                    )
                elif fields[1] == "f":
                    (idx, child) = self.parse_person(
                        fields, 2, Person.FEMALE, father_surname
                    )
                else:
                    (idx, child) = self.parse_person(
                        fields, 1, Person.UNKNOWN, father_surname
                    )

                if child:
                    childref = ChildRef()
                    childref.set_reference_handle(child.get_handle())
                    self.current_family.add_child_ref(childref)
                    self.db.commit_family(self.current_family, self.trans)
                    child.add_parent_family_handle(self.current_family.get_handle())
                    if self.current_child_birthplace_handle:
                        birth = None
                        birth_ref = child.get_birth_ref()
                        if birth_ref:
                            birth = self.db.get_event_from_handle(birth_ref.ref)
                        if not birth:
                            birth = self.create_event(EventType.BIRTH)
                            birth_ref = EventRef()
                            birth_ref.set_reference_handle(birth.get_handle())
                            child.set_birth_ref(birth_ref)
                        birth.set_place_handle(self.current_child_birthplace_handle)
                        self.db.commit_event(birth, self.trans)
                    if self.current_child_source_handle:
                        child.add_citation(self.current_child_source_handle)
                    self.db.commit_person(child, self.trans)
            else:
                break
        self.current_mode = None
        return None

    def read_children_birthplace_line(self, line, fields):
        cbp = self.get_or_create_place(self.decode(fields[1]))
        if cbp:
            self.current_child_birthplace_handle = cbp.get_handle()
        return None

    def read_children_source_line(self, line, fields):
        csrc = self.get_or_create_source(self.decode(fields[1]))
        self.current_child_source_handle = csrc.handle
        return None

    def read_family_comment(self, line, fields):
        if not self.current_family:
            LOG.warning("Unknown family of child in line %d!" % self.lineno)
            return None
        n = Note()
        n.set(line)
        self.db.add_note(n, self.trans)
        self.current_family.add_note(n.handle)
        self.db.commit_family(self.current_family, self.trans)
        return None

    def _read_notes_lines(self, note_tag):
        note_txt = ""
        while True:
            line = self.get_next_line()
            if line is None:
                break

            fields = line.split(" ")
            if fields[0] == "end" and fields[1] == note_tag:
                break
            elif fields[0] == "beg":
                continue
            else:
                if note_txt:
                    note_txt = note_txt + "\n" + line
                else:
                    note_txt = note_txt + line
        if note_txt:
            n = Note()
            n.set(note_txt)
            self.db.add_note(n, self.trans)
            return n.handle
        return None

    def read_person_notes_lines(self, line, fields):
        (idx, person) = self.parse_person(fields, 1, None, None)
        note_handle = self._read_notes_lines(fields[0])
        if note_handle:
            person.add_note(note_handle)
            self.db.commit_person(person, self.trans)

    def read_database_notes_lines(self, line, fields):
        note_handle = self._read_notes_lines(fields[0])

    def parse_marriage(self, fields, idx):
        mariageDataRe = re.compile("^[+#-0-9].*$")

        mar_date = None
        mar_place = None
        mar_source = None

        sep_date = None
        div_date = None

        married = 1
        engaged = 0

        # skip to marriage date in case person contained unmatches tokens
        # Alex: this failed when fields[idx] was an empty line. Fixed.
        # while idx < len(fields) and not fields[idx][0] == "+":
        while idx < len(fields) and not (fields[idx] and fields[idx][0] == "+"):
            if fields[idx]:
                LOG.warning(
                    ("parse_marriage(): Unknown field: " + "'%s' in line %d!")
                    % (fields[idx], self.lineno)
                )
            idx += 1

        while idx < len(fields) and mariageDataRe.match(fields[idx]):
            field = fields[idx]
            idx += 1
            if field.startswith("+"):
                field = field[1:]
                mar_date = self.parse_date(self.decode(field))
                LOG.debug(" Married at: %s" % field)
            elif field.startswith("-"):
                field = field[1:]
                div_date = self.parse_date(self.decode(field))
                LOG.debug(" Div at: %s" % field)
            elif field == "#mp" and idx < len(fields):
                mar_place = self.get_or_create_place(self.decode(fields[idx]))
                LOG.debug(" Marriage place: %s" % fields[idx])
                idx += 1
            elif field == "#ms" and idx < len(fields):
                mar_source = self.get_or_create_source(self.decode(fields[idx]))
                LOG.debug(" Marriage source: %s" % fields[idx])
                idx += 1
            elif field == "#sep" and idx < len(fields):
                sep_date = self.parse_date(self.decode(fields[idx]))
                LOG.debug(" Seperated since: %s" % fields[idx])
                idx += 1
            elif field == "#nm":
                LOG.debug(" Are not married.")
                married = 0
            elif field == "#noment":
                LOG.debug(" Not mentioned.")
            elif field == "#eng":
                LOG.debug(" Are engaged.")
                engaged = 1
            else:
                LOG.warning(
                    ("parse_marriage(): Unknown field " + "'%s'for mariage in line %d!")
                    % (field, self.lineno)
                )

        if mar_date or mar_place or mar_source:
            mar = self.create_event(
                EventType.MARRIAGE, None, mar_date, mar_place, mar_source
            )
            mar_ref = EventRef()
            mar_ref.set_reference_handle(mar.get_handle())
            mar_ref.set_role(EventRoleType.FAMILY)
            self.current_family.add_event_ref(mar_ref)
            self.current_family.set_relationship(FamilyRelType(FamilyRelType.MARRIED))

        if div_date:
            div = self.create_event(EventType.DIVORCE, None, div_date, None, None)
            div_ref = EventRef()
            div_ref.set_reference_handle(div.get_handle())
            div_ref.set_role(EventRoleType.FAMILY)
            self.current_family.add_event_ref(div_ref)

        if sep_date or engaged:
            sep = self.create_event(EventType.ENGAGEMENT, None, sep_date, None, None)
            sep_ref = EventRef()
            sep_ref.set_reference_handle(sep.get_handle())
            sep_ref.set_role(EventRoleType.FAMILY)
            self.current_family.add_event_ref(sep_ref)

        if not married:
            self.current_family.set_relationship(FamilyRelType(FamilyRelType.UNMARRIED))

        self.db.commit_family(self.current_family, self.trans)
        return idx

    def parse_person(self, fields, idx, gender, father_surname):
        if not father_surname:
            if not idx < len(fields):
                LOG.warning("Missing surname of person in line %d!" % self.lineno)
                surname = ""
            else:
                surname = self.decode(fields[idx])
            idx += 1
        else:
            surname = father_surname

        if not idx < len(fields):
            LOG.warning("Missing firstname of person in line %d!" % self.lineno)
            firstname = ""
        else:
            firstname = self.decode(fields[idx])
        idx += 1
        if idx < len(fields) and father_surname:
            noSurnameRe = re.compile(r"^[({\[~><?0-9#].*$")
            if not noSurnameRe.match(fields[idx]):
                surname = self.decode(fields[idx])
                idx += 1

        LOG.debug("Person: %s %s" % (firstname, surname))
        person = self.get_or_create_person(firstname, surname)
        name = Name()
        name.set_type(NameType(NameType.BIRTH))
        name.set_first_name(firstname)
        surname_obj = name.get_primary_surname()
        surname_obj.set_surname(surname)
        person.set_primary_name(name)
        if person.get_gender() == Person.UNKNOWN and gender is not None:
            person.set_gender(gender)
        self.db.commit_person(person, self.trans)
        personDataRe = re.compile(r"^[kmes0-9<>~#\[({!].*$")
        dateRe = re.compile("^[kmes0-9~<>?]+.*$")

        source = None
        birth_parsed = False
        birth_date = None
        birth_place = None
        birth_source = None

        bapt_date = None
        bapt_place = None
        bapt_source = None

        death_date = None
        death_place = None
        death_source = None
        death_cause = None

        crem_date = None
        bur_date = None
        bur_place = None
        bur_source = None

        public_name = None
        firstname_aliases = []
        nick_names = []
        name_aliases = []
        surname_aliases = []

        while idx < len(fields) and personDataRe.match(fields[idx]):
            field = fields[idx]
            idx += 1
            if field.startswith("("):
                LOG.debug("Public Name: %s" % field)
                public_name = self.decode(field[1:-1])
            elif field.startswith("{"):
                LOG.debug("Firstsname Alias: %s" % field)
                firstname_aliases.append(self.decode(field[1:-1]))
            elif field.startswith("["):
                LOG.debug("Title: %s" % field)
                titleparts = self.decode(field[1:-1]).split(":")
                tname = ttitle = tplace = tstart = tend = tnth = None
                try:
                    tname = titleparts[0]
                    ttitle = titleparts[1]
                    if titleparts[2]:
                        tplace = self.get_or_create_place(titleparts[2])
                    tstart = self.parse_date(titleparts[3])
                    tend = self.parse_date(titleparts[4])
                    tnth = titleparts[5]
                except IndexError:  # not all parts are written all the time
                    pass
                if tnth:  # Append title numer to title
                    # TODO for Arabic, should the next comma be translated?
                    ttitle += ", " + tnth
                title = self.create_event(EventType.NOB_TITLE, ttitle, tstart, tplace)
                # TODO: Geneweb has a start date and an end date, and therefore
                # supports stuff like: FROM about 1955 TO between 1998 and 1999
                # gramps only supports one single date or range.
                if tname and tname != "*":
                    n = Note()
                    n.set(tname)
                    self.db.add_note(n, self.trans)
                    title.add_note(n.handle)
                title_ref = EventRef()
                title_ref.set_reference_handle(title.get_handle())
                person.add_event_ref(title_ref)
            elif field == "#nick" and idx < len(fields):
                LOG.debug("Nick Name: %s" % fields[idx])
                nick_names.append(self.decode(fields[idx]))
                idx += 1
            elif field == "#occu" and idx < len(fields):
                LOG.debug("Occupation: %s" % fields[idx])
                occu = self.create_event(EventType.OCCUPATION, self.decode(fields[idx]))
                occu_ref = EventRef()
                occu_ref.set_reference_handle(occu.get_handle())
                person.add_event_ref(occu_ref)
                idx += 1
            elif field == "#alias" and idx < len(fields):
                LOG.debug("Name Alias: %s" % fields[idx])
                name_aliases.append(self.decode(fields[idx]))
                idx += 1
            elif field == "#salias" and idx < len(fields):
                LOG.debug("Surname Alias: %s" % fields[idx])
                surname_aliases.append(self.decode(fields[idx]))
                idx += 1
            elif field == "#image" and idx < len(fields):
                LOG.debug("Image: %s" % fields[idx])
                idx += 1
            elif field == "#src" and idx < len(fields):
                LOG.debug("Source: %s" % fields[idx])
                source = self.get_or_create_source(self.decode(fields[idx]))
                idx += 1
            elif field == "#bs" and idx < len(fields):
                LOG.debug("Birth Source: %s" % fields[idx])
                birth_source = self.get_or_create_source(self.decode(fields[idx]))
                idx += 1
            elif field[0] == "!":
                LOG.debug("Baptize at: %s" % field[1:])
                bapt_date = self.parse_date(self.decode(field[1:]))
            elif field == "#bp" and idx < len(fields):
                LOG.debug("Birth Place: %s" % fields[idx])
                birth_place = self.get_or_create_place(self.decode(fields[idx]))
                idx += 1
            elif field == "#pp" and idx < len(fields):
                LOG.debug("Baptize Place: %s" % fields[idx])
                bapt_place = self.get_or_create_place(self.decode(fields[idx]))
                idx += 1
            elif field == "#ps" and idx < len(fields):
                LOG.debug("Baptize Source: %s" % fields[idx])
                bapt_source = self.get_or_create_source(self.decode(fields[idx]))
                idx += 1
            elif field == "#dp" and idx < len(fields):
                LOG.debug("Death Place: %s" % fields[idx])
                death_place = self.get_or_create_place(self.decode(fields[idx]))
                idx += 1
            elif field == "#ds" and idx < len(fields):
                LOG.debug("Death Source: %s" % fields[idx])
                death_source = self.get_or_create_source(self.decode(fields[idx]))
                idx += 1
            elif field == "#buri" and idx < len(fields):
                if fields[idx][0] != "#":  # bug in GeneWeb: empty #buri fields
                    LOG.debug("Burial Date: %s" % fields[idx])
                    bur_date = self.parse_date(self.decode(fields[idx]))
                    idx += 1
            elif field == "#crem" and idx < len(fields):
                LOG.debug("Cremention Date: %s" % fields[idx])
                crem_date = self.parse_date(self.decode(fields[idx]))
                idx += 1
            elif field == "#rp" and idx < len(fields):
                LOG.debug("Burial Place: %s" % fields[idx])
                bur_place = self.get_or_create_place(self.decode(fields[idx]))
                idx += 1
            elif field == "#rs" and idx < len(fields):
                LOG.debug("Burial Source: %s" % fields[idx])
                bur_source = self.get_or_create_source(self.decode(fields[idx]))
                idx += 1
            elif field == "#apubl":
                LOG.debug("This is a public record")
            elif field == "#apriv":
                LOG.debug("This is a private record")
                person.set_privacy(True)
            elif field == "#h":
                LOG.debug("This is a restricted record")
                # TODO: Gramps does currently not feature this level
                person.set_privacy(True)
            elif dateRe.match(field):
                if not birth_parsed:
                    LOG.debug("Birth Date: %s" % field)
                    birth_date = self.parse_date(self.decode(field))
                    birth_parsed = True
                else:
                    LOG.debug("Death Date: %s" % field)
                    death_date = self.parse_date(self.decode(field))
                    if field == "mj":
                        death_cause = "Died joung"
                    elif field.startswith("k"):
                        death_cause = "Killed"
                    elif field.startswith("m"):
                        death_cause = "Murdered"
                    elif field.startswith("e"):
                        death_cause = "Executed"
                    elif field.startswith("d"):
                        death_cause = "Disappeared"
                    # TODO: Set special death types more properly
            else:
                LOG.warning(
                    ("parse_person(): Unknown field " + "'%s' for person in line %d!")
                    % (field, self.lineno)
                )

        if public_name:
            name = person.get_primary_name()
            name.set_type(NameType(NameType.BIRTH))
            person.add_alternate_name(name)
            name = Name()
            name.set_type(NameType(NameType.AKA))
            name.set_first_name(public_name)
            surname_obj = name.get_primary_surname()
            surname_obj.set_surname(surname)
            person.set_primary_name(name)

        for aka in nick_names:
            name = Attribute()
            name.set_type(AttributeType(AttributeType.NICKNAME))
            name.set_value(aka)
            person.add_attribute(name)

        for aka in firstname_aliases:
            name = Name()
            name.set_type(NameType(NameType.AKA))
            name.set_first_name(aka)
            surname_obj = name.get_primary_surname()
            surname_obj.set_surname(surname)
            person.add_alternate_name(name)

        for aka in name_aliases:
            name = Name()
            name.set_type(NameType(NameType.AKA))
            name.set_first_name(aka)
            surname_obj = name.get_primary_surname()
            surname_obj.set_surname(surname)
            person.add_alternate_name(name)

        for aka in surname_aliases:
            name = Name()
            name.set_type(NameType(NameType.AKA))
            if public_name:
                name.set_first_name(public_name)
            else:
                name.set_first_name(firstname)
            surname_obj = name.get_primary_surname()
            surname_obj.set_surname(aka)
            person.add_alternate_name(name)

        if source:
            person.add_citation(source.get_handle())

        if birth_date or birth_place or birth_source:
            birth = self.create_event(
                EventType.BIRTH, None, birth_date, birth_place, birth_source
            )
            birth_ref = EventRef()
            birth_ref.set_reference_handle(birth.get_handle())
            person.set_birth_ref(birth_ref)

        if bapt_date or bapt_place or bapt_source:
            babt = self.create_event(
                EventType.BAPTISM, None, bapt_date, bapt_place, bapt_source
            )
            babt_ref = EventRef()
            babt_ref.set_reference_handle(babt.get_handle())
            person.add_event_ref(babt_ref)

        if death_date or death_place or death_source or death_cause:
            death = self.create_event(
                EventType.DEATH, None, death_date, death_place, death_source
            )
            if death_cause:
                death.set_description(death_cause)
                self.db.commit_event(death, self.trans)
            death_ref = EventRef()
            death_ref.set_reference_handle(death.get_handle())
            person.set_death_ref(death_ref)

        if bur_date:
            bur = self.create_event(
                EventType.BURIAL, None, bur_date, bur_place, bur_source
            )
            bur_ref = EventRef()
            bur_ref.set_reference_handle(bur.get_handle())
            person.add_event_ref(bur_ref)

        if crem_date:
            crem = self.create_event(
                EventType.CREMATION, None, crem_date, bur_place, bur_source
            )
            crem_ref = EventRef()
            crem_ref.set_reference_handle(crem.get_handle())
            person.add_event_ref(crem_ref)

        self.db.commit_person(person, self.trans)

        return (idx, person)

    def parse_date(self, field):
        if field == "0":
            return None
        date = Date()
        matches = _text_parse.match(field)
        if matches:
            groups = matches.groups()
            date.set_as_text(groups[0])
            date.set_modifier(Date.MOD_TEXTONLY)
            return date

        matches = _date_parse.match(field)
        if matches:
            groups = matches.groups()
            mod = _mod_map.get(groups[0], Date.MOD_NONE)
            if groups[3] == "..":
                if groups[4] == "0":
                    mod = Date.MOD_FROM
                    cal1 = _cal_map.get(groups[2], Date.CAL_GREGORIAN)
                    sub1 = self.sub_date(groups[1])
                    sub2 = (0, 0, 0)
                elif groups[1] == "0":
                    mod = Date.MOD_TO
                    cal1 = _cal_map.get(groups[5], Date.CAL_GREGORIAN)
                    sub1 = self.sub_date(groups[4])
                    sub2 = (0, 0, 0)
                else:
                    mod = Date.MOD_SPAN
                    cal1 = _cal_map.get(groups[2], Date.CAL_GREGORIAN)
                    cal2 = _cal_map.get(groups[5], Date.CAL_GREGORIAN)
                    sub1 = self.sub_date(groups[1])
                    sub2 = self.sub_date(groups[4])
            else:
                cal1 = _cal_map.get(groups[2], Date.CAL_GREGORIAN)
                sub1 = self.sub_date(groups[1])
                sub2 = (0, 0, 0)
            try:
                date.set(
                    Date.QUAL_NONE,
                    mod,
                    cal1,
                    (sub1[0], sub1[1], sub1[2], 0, sub2[0], sub2[1], sub2[2], 0),
                )
            except DateError as e:
                # Translators: leave the {date} and {gw_snippet} untranslated
                # in the format string, but you may re-order them if needed.
                LOG.warning(
                    _(
                        "Invalid date {date} in {gw_snippet}, "
                        "preserving date as text."
                    ).format(date=e.date.__dict__, gw_snippet=field)
                )
                date.set(modifier=Date.MOD_TEXTONLY, text=field)
            return date
        else:
            return None

    def sub_date(self, data):
        vals = data.split("/")
        if len(vals) == 1:
            return (0, 0, int(vals[0]))
        elif len(vals) == 2:
            return (0, int(vals[0]), int(vals[1]))
        else:
            return (int(vals[0]), int(vals[1]), int(vals[2]))

    def create_event(self, type, desc=None, date=None, place=None, source=None):
        event = Event()
        if type:
            event.set_type(EventType(type))
        if desc:
            event.set_description(desc)
        if date:
            event.set_date_object(date)
        if place:
            event.set_place_handle(place.get_handle())
        if source:
            event.add_citation(source.get_handle())
        self.db.add_event(event, self.trans)
        self.db.commit_event(event, self.trans)
        return event

    def get_or_create_person(self, firstname, lastname):
        person = None
        mykey = firstname + lastname
        if mykey in self.ikeys and firstname != "?" and lastname != "?":
            person = self.db.get_person_from_handle(self.ikeys[mykey])
        else:
            person = Person()
            self.db.add_person(person, self.trans)
            self.db.commit_person(person, self.trans)
            self.ikeys[mykey] = person.get_handle()
        return person

    def get_or_create_place(self, place_name):
        place = None
        if place_name in self.pkeys:
            place = self.db.get_place_from_handle(self.pkeys[place_name])
        else:
            place = Place()
            place.name = PlaceName(value=place_name)
            place.set_title(place_name)
            self.db.add_place(place, self.trans)
            self.db.commit_place(place, self.trans)
            self.pkeys[place_name] = place.get_handle()
        return place

    def get_or_create_source(self, source_name):
        source = None
        if source_name in self.skeys:
            source = self.db.get_source_from_handle(self.skeys[source_name])
        else:
            source = Source()
            source.set_title(source_name)
            self.db.add_source(source, self.trans)
            self.db.commit_source(source, self.trans)
            self.skeys[source_name] = source.get_handle()
        citation = Citation()
        citation.set_reference_handle(source.get_handle())
        self.db.add_citation(citation, self.trans)
        self.db.commit_citation(citation, self.trans)
        return citation

    def read_fevent_line(self, event):
        if fevents_map.get(event[0:5]) == None:
            return  # need to fix custom event types not in the map

        fev = None
        # get events for the current family
        for evr in self.current_family.get_event_ref_list():
            ev = self.db.get_event_from_handle(evr.get_reference_handle())
            if ev.get_type() == fevents_map.get(event[0:5]):
                fev = ev  # found. Need to also check EventRef role
                return
            if not fev:  # No event found create a new one
                if evr.get_role() != EventRoleType(EventRoleType.FAMILY):
                    continue
                else:
                    LOG.info((ev.get_type(), self.current_family.handle))
                    self.new_gwplus_fevent(event)
        while True:
            line = self.get_next_line()
            if line and line[0:5] in fevents_map:
                self.new_gwplus_fevent(line)
            elif line and line[0:4] == "wit:":
                continue
            else:
                self.current_mode = None
                # self.db.commit_family(self.current_family,self.trans)
                break

    def read_pevent_line(self, event, fields):
        name = fields[2] + fields[1]

        try:
            self.person = self.ikeys[name]
        # check key on {ikey}
        except:
            self.person = "(TO_CHECK: %s)" % fields[1:]
            # GrampsImportError()

        lastname = fields[1]
        firstname = fields[2]
        self.current_person = self.get_or_create_person(firstname, lastname)

        # name = Name()
        # name.set_type(NameType(NameType.BIRTH))
        # name.set_first_name(firstname)
        # surname_obj = name.get_primary_surname()
        # surname_obj.set_surname(surname)
        # self.current_person.set_primary_name(name)

        if pevents_map.get(event[0:5]) == None:
            return  # need to fix custom event types not in the map

        self.current_event = None
        # get events for the current person
        for evr in self.current_person.get_event_ref_list():
            ev = self.db.get_event_from_handle(evr.get_reference_handle())
            if ev.get_type() == pevents_map.get(event[0:5]):
                self.current_event = ev  # found. Need to also check EventRef role
            if not self.current_event:  # No event found create a new one
                self.current_event = self.new_gwplus_pevent(event)
        while True:
            line = self.get_next_line()
            if line and line[0:5] in pevents_map:
                self.current_mode = "person_event"
                self.current_event = self.new_gwplus_pevent(line)
            elif line and line[0:4] == "note":
                n = Note()
                n.set(line[5:])
                self.db.add_note(n, self.trans)
                if self.current_event:
                    self.current_event.add_note(n.handle)
                    self.db.commit_event(self.current_event, self.trans)
                else:
                    print("note", n.handle)
            else:
                self.current_mode = None
                # self.db.commit_person(self.current_person,self.trans)
                break

    def new_gwplus_fevent(self, line):
        source = place = note = type = None
        date = self.parse_date(self.decode(line[6:]))

        idx = 0
        LOG.info((line, fevents_map.get(line[0:5])))
        type = fevents_map.get(line[0:5])
        data = line.split()
        date = self.parse_date(self.decode(line[6:]))
        for part in data:
            idx += 1
            if part == "#p":
                place = self.get_or_create_place(self.decode(data[idx]))
            if part == "#s":
                source = self.get_or_create_source(self.decode(data[idx]))
        self.current_event = self.create_event(type, None, None, None, None)
        print("new event", self.current_event.handle)
        if date:
            print(date)
            self.current_event.set_date_object(date)
        if place:
            print("place", place.handle)
            self.current_event.set_place_handle(place.get_handle())
        if source:
            print("source", source.handle)
            self.current_event.add_citation(source.get_handle())
        self.db.commit_event(self.current_event, self.trans)
        nev_ref = EventRef()
        nev_ref.set_reference_handle(self.current_event.get_handle())
        self.current_family.add_event_ref(nev_ref)
        self.db.commit_family(self.current_family, self.trans)
        return self.current_event

    def new_gwplus_pevent(self, line):
        source = place = note = type = None
        date = self.parse_date(self.decode(line[6:]))

        idx = 0
        LOG.info((self.person, line, pevents_map.get(line[0:5])))
        type = pevents_map.get(line[0:5])
        data = line.split()
        date = self.parse_date(self.decode(line[6:]))
        for part in data:
            idx += 1
            if part == "#p":
                place = self.get_or_create_place(self.decode(data[idx]))
            if part == "#s":
                source = self.get_or_create_source(self.decode(data[idx]))
        self.current_event = self.create_event(type, None, None, None, None)
        print("new event", self.current_event.handle)
        if date:
            print(date)
            self.current_event.set_date_object(date)
        if place:
            print("place", place.handle)
            self.current_event.set_place_handle(place.get_handle())
        if source:
            print("source", source.handle)
            self.current_event.add_citation(source.get_handle())
        self.db.commit_event(self.current_event, self.trans)
        nev_ref = EventRef()
        nev_ref.set_reference_handle(self.current_event.get_handle())
        self.current_person.add_event_ref(nev_ref)
        self.db.commit_person(self.current_person, self.trans)
        return self.current_event

    def decode(self, s):
        s = s.replace("_", " ")
        charref_re = re.compile("(&#)(x?)([0-9a-zA-Z]+)(;)")
        for match in charref_re.finditer(s):
            try:
                if match.group(2):  # HEX
                    nchar = chr(int(match.group(3), 16))
                else:  # Decimal
                    nchar = chr(int(match.group(3)))
                s = s.replace(match.group(0), nchar)
            except UnicodeDecodeError:
                pass

        # replace named entities
        entref_re = re.compile("(&)([a-zA-Z]+)(;)")
        for match in entref_re.finditer(s):
            try:
                if match.group(2) in name2codepoint:
                    nchar = chr(name2codepoint[match.group(2)])
                s = s.replace(match.group(0), nchar)
            except UnicodeDecodeError:
                pass

        return s

    def debug(self, txt):
        LOG.debug(txt)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
# Copyright (C) 2018       Theo van Rijn
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    ContactPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.utils.config import get_researcher
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class ContactPage(BasePage):
    """
    This class is responsible for displaying information about the 'Researcher'
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class for
                             this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)

        output_file, sio = self.report.create_file("contact")
        result = self.write_header(self._("Contact"))
        contactpage, head, dummy_body, outerwrapper = result
        ldatec = 0

        # begin contact division
        with Html("div", class_="content", id="Contact") as section:
            outerwrapper += section

            # begin summaryarea division
            with Html("div", id="summaryarea") as summaryarea:
                section += summaryarea

                contactimg = self.add_image("contactimg", head, 200)
                if contactimg is not None:
                    summaryarea += contactimg

                # get researcher information
                res = get_researcher()

                with Html("div", id="researcher") as researcher:
                    summaryarea += researcher

                    if res.name:
                        res.name = res.name.replace(",,,", "")
                        researcher += Html("h3", res.name, inline=True)
                    if res.addr:
                        researcher += Html(
                            "span", res.addr, id="streetaddress", inline=True
                        )
                    if res.locality:
                        researcher += Html(
                            "span", res.locality, id="locality", inline=True
                        )
                    text = "".join([res.city, res.state, res.postal])
                    if text:
                        city = Html("span", res.city, id="city", inline=True)
                        state = Html("span", res.state, id="state", inline=True)
                        postal = Html("span", res.postal, id="postalcode", inline=True)
                        researcher += (city, state, postal)
                    if res.country:
                        researcher += Html(
                            "span", res.country, id="country", inline=True
                        )
                    if res.email:
                        researcher += Html("span", id="email") + (
                            Html(
                                "a",
                                res.email,
                                href="mailto:%s" % res.email,
                                inline=True,
                            )
                        )

                    # add clear line for proper styling
                    summaryarea += FULLCLEAR

                    note_id = report.options["contactnote"]
                    if note_id:
                        note = self.r_db.get_note_from_gramps_id(note_id)
                        if note:
                            note_text = self.get_note_format(note, False)

                            # attach note
                            summaryarea += note_text

                            # last modification of this note
                            ldatec = note.get_change_time()

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for porcessing
        # and close the file
        self.xhtml_writer(contactpage, output_file, sio, 0)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for config.py"""

import os
import unittest
from ..config import ConfigManager


def callback(*args):
    # args: self, 0, str(setting), None
    # print "Calling callback with:", args
    # self is ConfigManager
    self = args[0]
    self._x = args[2]


TEMP_INI = "./temp.ini"
TEST2_INI = "./test2.ini"


class CompleteCheck(unittest.TestCase):
    def tearDown(self):
        os.remove(TEMP_INI)
        os.remove(TEST2_INI)

    def testAll(self):
        self.CM = ConfigManager(TEMP_INI)
        self.CM.register("section.setting1", 1)  # int
        self.CM.register("section.setting2", 3.1415)  # float
        self.CM.register("section.setting3", "String")  # string
        self.CM.register("section.setting4", False)  # boolean

        self.assertEqual(self.CM.get("section.setting1"), 1)
        self.assertEqual(self.CM.get("section.setting2"), 3.1415)
        self.assertEqual(self.CM.get("section.setting3"), "String")
        self.assertFalse(self.CM.get("section.setting4"))

        self.CM.set("section.setting1", 2)
        self.CM.set("section.setting2", 8.6)
        self.CM.set("section.setting3", "Another String")
        self.CM.set("section.setting4", True)

        self.assertEqual(self.CM.get("section.setting1"), 2)
        self.assertEqual(self.CM.get("section.setting2"), 8.6)
        self.assertEqual(self.CM.get("section.setting3"), "Another String")
        self.assertTrue(self.CM.get("section.setting4"))

        # self.assertRaises(AttributeError, self.CM.set, "section.setting1", 2.8)
        # self.assertRaises(AttributeError, self.CM.set, "section.setting2", 2)
        self.assertRaises(AttributeError, self.CM.set, "section.setting3", 6)
        self.assertRaises(AttributeError, self.CM.set, "section.setting4", 1)

        self.assertEqual(self.CM.get("section.setting1"), 2)
        self.assertEqual(self.CM.get("section.setting2"), 8.6)
        self.assertEqual(self.CM.get("section.setting3"), "Another String")
        self.assertTrue(self.CM.get("section.setting4"))

        # Try to set one that doesn't exist:
        self.assertRaises(AttributeError, self.CM.set, "section.setting5", 1)

        self.CM.save()

        self.CM.reset()  # to defaults

        self.assertEqual(self.CM.get("section.setting1"), 1)
        self.assertEqual(
            self.CM.get("section.setting2"), 3.1415, self.CM.get("section.setting2")
        )
        self.assertEqual(self.CM.get("section.setting3"), "String")
        self.assertFalse(self.CM.get("section.setting4"))

        self.CM._x = None

        cbid = self.CM.connect("section.setting1", callback)
        self.assertIsNone(self.CM._x)

        self.CM.set("section.setting1", 1024)
        self.assertEqual(self.CM._x, "1024", "x was '%s'" % self.CM._x)

        self.CM.disconnect(cbid)

        self.CM.set("section.setting1", -1)
        self.assertEqual(self.CM._x, "1024")

        self.CM.reset("section.setting1")
        self.assertEqual(self.CM.get("section.setting1"), 1)

        # No callback:
        self.CM._x = None
        self.CM.set("section.setting1", 200)
        self.assertIsNone(self.CM._x)
        # Now, set one:
        cbid = self.CM.connect("section.setting1", callback)
        # Now, call it:
        self.CM.emit("section.setting1")
        self.assertEqual(self.CM._x, "200")

        self.CM.set("section.setting1", 2)
        self.CM.set("section.setting2", 8.6)
        self.CM.set("section.setting3", "Another String")
        self.CM.set("section.setting4", True)

        self.CM.register("section2.windows-file", r"c:\drive\path\o'malley\file.pdf")
        self.CM.register("section2.list", [1, 2, 3, 4])
        self.CM.register("section2.dict", {"a": "apple", "b": "banana"})
        self.CM.register("section2.unicode", "Raötröme")

        self.CM.save(TEST2_INI)
        self.CM.reset()

        self.assertEqual(
            self.CM.get("section.setting1"), 1, self.CM.get("section.setting1")
        )
        self.assertEqual(self.CM.get("section.setting2"), 3.1415)
        self.assertEqual(self.CM.get("section.setting3"), "String")
        self.assertFalse(self.CM.get("section.setting4"))

        self.CM.load(TEST2_INI)

        self.assertEqual(
            self.CM.get("section.setting1"), 2, self.CM.get("section.setting1")
        )
        self.assertEqual(self.CM.get("section.setting2"), 8.6)
        self.assertEqual(self.CM.get("section.setting3"), "Another String")
        self.assertTrue(self.CM.get("section.setting4"))

        self.assertEqual(
            self.CM.get("section2.windows-file"), r"c:\drive\path\o'malley\file.pdf"
        )
        self.assertEqual(self.CM.get("section2.list"), [1, 2, 3, 4])
        self.assertEqual(self.CM.get("section2.dict"), {"a": "apple", "b": "banana"})
        self.assertEqual(self.CM.get("section2.unicode"), "Raötröme")

        self.assertRaises(AttributeError, self.CM.save, TEST2_INI, comments=123)
        self.assertRaises(
            AttributeError, self.CM.save, TEST2_INI, comments={"key": "pair"}
        )
        self.assertRaises(AttributeError, self.CM.save, TEST2_INI, comments=[123])
        self.assertRaises(
            AttributeError, self.CM.save, TEST2_INI, comments=["line1", 123]
        )

        self.CM.save(TEST2_INI, comments=["test comment1"])
        self.CM.load(TEST2_INI)
        self.assertEqual(self.CM.get("section.setting3"), "Another String")

        self.CM.save(TEST2_INI, comments=["test comment1", "test comment2"])
        self.CM.load(TEST2_INI)
        self.assertEqual(self.CM.get("section.setting3"), "Another String")


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016      Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import unittest

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.db import DbTxn
from gramps.gen.db.utils import make_database
from gramps.gen.lib import (
    Person,
    Family,
    Event,
    Place,
    Repository,
    Source,
    Citation,
    Media,
    Note,
    Tag,
    Researcher,
    Surname,
)


# -------------------------------------------------------------------------
#
# DbRandomTest class
#
# -------------------------------------------------------------------------
class DbRandomTest(unittest.TestCase):
    """
    Tests with random objects.
    """

    @classmethod
    def setUpClass(cls):
        cls.db = make_database("sqlite")
        cls.db.load(":memory:")

    def setUp(self):
        self.handles = {
            "Person": [],
            "Family": [],
            "Event": [],
            "Place": [],
            "Repository": [],
            "Source": [],
            "Citation": [],
            "Media": [],
            "Note": [],
            "Tag": [],
        }
        self.gids = {
            "Person": [],
            "Family": [],
            "Event": [],
            "Place": [],
            "Repository": [],
            "Source": [],
            "Citation": [],
            "Media": [],
            "Note": [],
        }
        with DbTxn("Add test objects", self.db) as trans:
            for i in range(10):
                self.__add_object(Person, self.db.add_person, trans)
                self.__add_object(Family, self.db.add_family, trans)
                self.__add_object(Event, self.db.add_event, trans)
                self.__add_object(Place, self.db.add_place, trans)
                self.__add_object(Repository, self.db.add_repository, trans)
                self.__add_object(Source, self.db.add_source, trans)
                self.__add_object(Citation, self.db.add_citation, trans)
                self.__add_object(Media, self.db.add_media, trans)
                self.__add_object(Note, self.db.add_note, trans)
                self.__add_object(Tag, self.db.add_tag, trans)

    def tearDown(self):
        with DbTxn("Remove test objects", self.db) as trans:
            for handle in self.handles["Person"]:
                self.db.remove_person(handle, trans)
            for handle in self.handles["Family"]:
                self.db.remove_family(handle, trans)
            for handle in self.handles["Event"]:
                self.db.remove_event(handle, trans)
            for handle in self.handles["Place"]:
                self.db.remove_place(handle, trans)
            for handle in self.handles["Repository"]:
                self.db.remove_repository(handle, trans)
            for handle in self.handles["Source"]:
                self.db.remove_source(handle, trans)
            for handle in self.handles["Citation"]:
                self.db.remove_citation(handle, trans)
            for handle in self.handles["Media"]:
                self.db.remove_media(handle, trans)
            for handle in self.handles["Note"]:
                self.db.remove_note(handle, trans)
            for handle in self.handles["Tag"]:
                self.db.remove_tag(handle, trans)

    def __add_object(self, obj_class, add_func, trans):
        obj = obj_class()
        if obj_class == Tag:
            obj.name = "Tag%s" % len(self.handles["Tag"])
        handle = add_func(obj, trans)
        self.handles[obj_class.__name__].append(handle)
        if obj_class != Tag:
            self.gids[obj_class.__name__].append(obj.gramps_id)

    ################################################################
    #
    # Test get_number_of_* methods
    #
    ################################################################

    def test_number_of_people(self):
        self.assertEqual(self.db.get_number_of_people(), len(self.handles["Person"]))

    def test_number_of_families(self):
        self.assertEqual(self.db.get_number_of_families(), len(self.handles["Family"]))

    def test_number_of_events(self):
        self.assertEqual(self.db.get_number_of_events(), len(self.handles["Event"]))

    def test_number_of_places(self):
        self.assertEqual(self.db.get_number_of_places(), len(self.handles["Place"]))

    def test_number_of_repositories(self):
        self.assertEqual(
            self.db.get_number_of_repositories(), len(self.handles["Repository"])
        )

    def test_number_of_sources(self):
        self.assertEqual(self.db.get_number_of_sources(), len(self.handles["Source"]))

    def test_number_of_citations(self):
        self.assertEqual(
            self.db.get_number_of_citations(), len(self.handles["Citation"])
        )

    def test_number_of_media(self):
        self.assertEqual(self.db.get_number_of_media(), len(self.handles["Media"]))

    def test_number_of_notes(self):
        self.assertEqual(self.db.get_number_of_notes(), len(self.handles["Note"]))

    def test_number_of_tags(self):
        self.assertEqual(self.db.get_number_of_tags(), len(self.handles["Tag"]))

    ################################################################
    #
    # Test get_*_handles methods
    #
    ################################################################
    def __get_handles_test(
        self, obj_type, handles_func, number_func, sort_handles=False
    ):
        if sort_handles:
            handles = handles_func(sort_handles=True)
        else:
            handles = handles_func()
        self.assertEqual(len(handles), number_func())
        for handle in handles:
            self.assertIn(handle, self.handles[obj_type])

    def test_get_person_handles(self):
        self.__get_handles_test(
            "Person", self.db.get_person_handles, self.db.get_number_of_people
        )

    def test_get_family_handles(self):
        self.__get_handles_test(
            "Family", self.db.get_family_handles, self.db.get_number_of_families
        )

    def test_get_event_handles(self):
        self.__get_handles_test(
            "Event", self.db.get_event_handles, self.db.get_number_of_events
        )

    def test_get_place_handles(self):
        self.__get_handles_test(
            "Place", self.db.get_place_handles, self.db.get_number_of_places
        )

    def test_get_repository_handles(self):
        self.__get_handles_test(
            "Repository",
            self.db.get_repository_handles,
            self.db.get_number_of_repositories,
        )

    def test_get_source_handles(self):
        self.__get_handles_test(
            "Source", self.db.get_source_handles, self.db.get_number_of_sources
        )

    def test_get_citation_handles(self):
        self.__get_handles_test(
            "Citation", self.db.get_citation_handles, self.db.get_number_of_citations
        )

    def test_get_media_handles(self):
        self.__get_handles_test(
            "Media", self.db.get_media_handles, self.db.get_number_of_media
        )

    def test_get_note_handles(self):
        self.__get_handles_test(
            "Note", self.db.get_note_handles, self.db.get_number_of_notes
        )

    def test_get_tag_handles(self):
        self.__get_handles_test(
            "Tag", self.db.get_tag_handles, self.db.get_number_of_tags
        )

    def test_get_person_handles_sort(self):
        self.__get_handles_test(
            "Person",
            self.db.get_person_handles,
            self.db.get_number_of_people,
            sort_handles=True,
        )

    def test_get_family_handles_sort(self):
        self.__get_handles_test(
            "Family",
            self.db.get_family_handles,
            self.db.get_number_of_families,
            sort_handles=True,
        )

    def test_get_place_handles_sort(self):
        self.__get_handles_test(
            "Place",
            self.db.get_place_handles,
            self.db.get_number_of_places,
            sort_handles=True,
        )

    def test_get_source_handles_sort(self):
        self.__get_handles_test(
            "Source",
            self.db.get_source_handles,
            self.db.get_number_of_sources,
            sort_handles=True,
        )

    def test_get_citation_handles_sort(self):
        self.__get_handles_test(
            "Citation",
            self.db.get_citation_handles,
            self.db.get_number_of_citations,
            sort_handles=True,
        )

    def test_get_media_handles_sort(self):
        self.__get_handles_test(
            "Media",
            self.db.get_media_handles,
            self.db.get_number_of_media,
            sort_handles=True,
        )

    def test_get_tag_handles_sort(self):
        self.__get_handles_test(
            "Tag",
            self.db.get_tag_handles,
            self.db.get_number_of_tags,
            sort_handles=True,
        )

    ################################################################
    #
    # Test get_*_gramps_ids methods
    #
    ################################################################
    def __get_gids_test(self, obj_type, gids_func, number_func):
        gids = gids_func()
        self.assertEqual(len(gids), number_func())
        for gid in gids:
            self.assertIn(gid, self.gids[obj_type])

    def test_get_person_gids(self):
        self.__get_gids_test(
            "Person", self.db.get_person_gramps_ids, self.db.get_number_of_people
        )

    def test_get_family_gids(self):
        self.__get_gids_test(
            "Family", self.db.get_family_gramps_ids, self.db.get_number_of_families
        )

    def test_get_event_gids(self):
        self.__get_gids_test(
            "Event", self.db.get_event_gramps_ids, self.db.get_number_of_events
        )

    def test_get_place_gids(self):
        self.__get_gids_test(
            "Place", self.db.get_place_gramps_ids, self.db.get_number_of_places
        )

    def test_get_repository_gids(self):
        self.__get_gids_test(
            "Repository",
            self.db.get_repository_gramps_ids,
            self.db.get_number_of_repositories,
        )

    def test_get_source_gids(self):
        self.__get_gids_test(
            "Source", self.db.get_source_gramps_ids, self.db.get_number_of_sources
        )

    def test_get_citation_gids(self):
        self.__get_gids_test(
            "Citation", self.db.get_citation_gramps_ids, self.db.get_number_of_citations
        )

    def test_get_media_gids(self):
        self.__get_gids_test(
            "Media", self.db.get_media_gramps_ids, self.db.get_number_of_media
        )

    def test_get_note_gids(self):
        self.__get_gids_test(
            "Note", self.db.get_note_gramps_ids, self.db.get_number_of_notes
        )

    ################################################################
    #
    # Test get_*_from_handle methods
    #
    ################################################################

    def __get_from_handle_test(self, obj_class, handles_func, get_func):
        for handle in handles_func():
            person = get_func(handle)
            self.assertIsInstance(person, obj_class)
            self.assertEqual(person.handle, handle)

    def test_get_person_from_handle(self):
        self.__get_from_handle_test(
            Person, self.db.get_person_handles, self.db.get_person_from_handle
        )

    def test_get_family_from_handle(self):
        self.__get_from_handle_test(
            Family, self.db.get_family_handles, self.db.get_family_from_handle
        )

    def test_get_event_from_handle(self):
        self.__get_from_handle_test(
            Event, self.db.get_event_handles, self.db.get_event_from_handle
        )

    def test_get_place_from_handle(self):
        self.__get_from_handle_test(
            Place, self.db.get_place_handles, self.db.get_place_from_handle
        )

    def test_get_repository_from_handle(self):
        self.__get_from_handle_test(
            Repository,
            self.db.get_repository_handles,
            self.db.get_repository_from_handle,
        )

    def test_get_source_from_handle(self):
        self.__get_from_handle_test(
            Source, self.db.get_source_handles, self.db.get_source_from_handle
        )

    def test_get_citation_from_handle(self):
        self.__get_from_handle_test(
            Citation, self.db.get_citation_handles, self.db.get_citation_from_handle
        )

    def test_get_media_from_handle(self):
        self.__get_from_handle_test(
            Media, self.db.get_media_handles, self.db.get_media_from_handle
        )

    def test_get_note_from_handle(self):
        self.__get_from_handle_test(
            Note, self.db.get_note_handles, self.db.get_note_from_handle
        )

    def test_get_tag_from_handle(self):
        self.__get_from_handle_test(
            Tag, self.db.get_tag_handles, self.db.get_tag_from_handle
        )

    ################################################################
    #
    # Test get_*_from_gramps_id methods
    #
    ################################################################

    def __get_from_gid_test(self, obj_class, gids_func, get_func):
        for gid in gids_func():
            person = get_func(gid)
            self.assertIsInstance(person, obj_class)
            self.assertEqual(person.gramps_id, gid)

    def test_get_person_from_gid(self):
        self.__get_from_gid_test(
            Person, self.db.get_person_gramps_ids, self.db.get_person_from_gramps_id
        )

    def test_get_family_from_gid(self):
        self.__get_from_gid_test(
            Family, self.db.get_family_gramps_ids, self.db.get_family_from_gramps_id
        )

    def test_get_event_from_gid(self):
        self.__get_from_gid_test(
            Event, self.db.get_event_gramps_ids, self.db.get_event_from_gramps_id
        )

    def test_get_place_from_gid(self):
        self.__get_from_gid_test(
            Place, self.db.get_place_gramps_ids, self.db.get_place_from_gramps_id
        )

    def test_get_repository_from_gid(self):
        self.__get_from_gid_test(
            Repository,
            self.db.get_repository_gramps_ids,
            self.db.get_repository_from_gramps_id,
        )

    def test_get_source_from_gid(self):
        self.__get_from_gid_test(
            Source, self.db.get_source_gramps_ids, self.db.get_source_from_gramps_id
        )

    def test_get_citation_from_gid(self):
        self.__get_from_gid_test(
            Citation,
            self.db.get_citation_gramps_ids,
            self.db.get_citation_from_gramps_id,
        )

    def test_get_media_from_gid(self):
        self.__get_from_gid_test(
            Media, self.db.get_media_gramps_ids, self.db.get_media_from_gramps_id
        )

    def test_get_note_from_gid(self):
        self.__get_from_gid_test(
            Note, self.db.get_note_gramps_ids, self.db.get_note_from_gramps_id
        )

    def test_get_tag_from_name(self):
        tag = self.db.get_tag_from_name("Tag0")
        self.assertEqual(tag.handle, self.handles["Tag"][0])

    ################################################################
    #
    # Test has_*_handle methods
    #
    ################################################################
    def test_has_person_handle(self):
        for handle in self.handles["Person"]:
            self.assertTrue(self.db.has_person_handle(handle))

    def test_has_family_handle(self):
        for handle in self.handles["Family"]:
            self.assertTrue(self.db.has_family_handle(handle))

    def test_has_event_handle(self):
        for handle in self.handles["Event"]:
            self.assertTrue(self.db.has_event_handle(handle))

    def test_has_place_handle(self):
        for handle in self.handles["Place"]:
            self.assertTrue(self.db.has_place_handle(handle))

    def test_has_repository_handle(self):
        for handle in self.handles["Repository"]:
            self.assertTrue(self.db.has_repository_handle(handle))

    def test_has_source_handle(self):
        for handle in self.handles["Source"]:
            self.assertTrue(self.db.has_source_handle(handle))

    def test_has_citation_handle(self):
        for handle in self.handles["Citation"]:
            self.assertTrue(self.db.has_citation_handle(handle))

    def test_has_media_handle(self):
        for handle in self.handles["Media"]:
            self.assertTrue(self.db.has_media_handle(handle))

    def test_has_note_handle(self):
        for handle in self.handles["Note"]:
            self.assertTrue(self.db.has_note_handle(handle))

    def test_has_tag_handle(self):
        for handle in self.handles["Tag"]:
            self.assertTrue(self.db.has_tag_handle(handle))

    ################################################################
    #
    # Test has_*_gramps_id methods
    #
    ################################################################
    def test_has_person_gramps_id(self):
        for gramps_id in self.gids["Person"]:
            self.assertTrue(self.db.has_person_gramps_id(gramps_id))

    def test_has_family_gramps_id(self):
        for gramps_id in self.gids["Family"]:
            self.assertTrue(self.db.has_family_gramps_id(gramps_id))

    def test_has_event_gramps_id(self):
        for gramps_id in self.gids["Event"]:
            self.assertTrue(self.db.has_event_gramps_id(gramps_id))

    def test_has_place_gramps_id(self):
        for gramps_id in self.gids["Place"]:
            self.assertTrue(self.db.has_place_gramps_id(gramps_id))

    def test_has_repository_gramps_id(self):
        for gramps_id in self.gids["Repository"]:
            self.assertTrue(self.db.has_repository_gramps_id(gramps_id))

    def test_has_source_gramps_id(self):
        for gramps_id in self.gids["Source"]:
            self.assertTrue(self.db.has_source_gramps_id(gramps_id))

    def test_has_citation_gramps_id(self):
        for gramps_id in self.gids["Citation"]:
            self.assertTrue(self.db.has_citation_gramps_id(gramps_id))

    def test_has_media_gramps_id(self):
        for gramps_id in self.gids["Media"]:
            self.assertTrue(self.db.has_media_gramps_id(gramps_id))

    def test_has_note_gramps_id(self):
        for gramps_id in self.gids["Note"]:
            self.assertTrue(self.db.has_note_gramps_id(gramps_id))

    ################################################################
    #
    # Test get_*_cursor methods
    #
    ################################################################
    def __get_cursor_test(self, cursor_func, raw_func):
        with cursor_func() as cursor:
            for handle, data1 in cursor:
                data2 = raw_func(handle)
                if isinstance(data1, dict) and isinstance(data2, dict):
                    self.assertDictEqual(dict(data1), dict(data2))
                else:
                    self.assertEqual(data1, data2)

    def test_get_person_cursor(self):
        self.__get_cursor_test(self.db.get_person_cursor, self.db.get_raw_person_data)

    def test_get_family_cursor(self):
        self.__get_cursor_test(self.db.get_family_cursor, self.db.get_raw_family_data)

    def test_get_event_cursor(self):
        self.__get_cursor_test(self.db.get_event_cursor, self.db.get_raw_event_data)

    def test_get_place_cursor(self):
        self.__get_cursor_test(self.db.get_place_cursor, self.db.get_raw_place_data)

    def test_get_repository_cursor(self):
        self.__get_cursor_test(
            self.db.get_repository_cursor, self.db.get_raw_repository_data
        )

    def test_get_source_cursor(self):
        self.__get_cursor_test(self.db.get_source_cursor, self.db.get_raw_source_data)

    def test_get_citation_cursor(self):
        self.__get_cursor_test(
            self.db.get_citation_cursor, self.db.get_raw_citation_data
        )

    def test_get_media_cursor(self):
        self.__get_cursor_test(self.db.get_media_cursor, self.db.get_raw_media_data)

    def test_get_note_cursor(self):
        self.__get_cursor_test(self.db.get_note_cursor, self.db.get_raw_note_data)

    def test_get_tag_cursor(self):
        self.__get_cursor_test(self.db.get_tag_cursor, self.db.get_raw_tag_data)

    ################################################################
    #
    # Test iter_*_handles methods
    #
    ################################################################

    def __iter_handles_test(self, obj_type, iter_func):
        for handle in iter_func():
            self.assertIn(handle, self.handles[obj_type])

    def test_iter_person_handles(self):
        self.__iter_handles_test("Person", self.db.iter_person_handles)

    def test_iter_family_handles(self):
        self.__iter_handles_test("Family", self.db.iter_family_handles)

    def test_iter_event_handles(self):
        self.__iter_handles_test("Event", self.db.iter_event_handles)

    def test_iter_place_handles(self):
        self.__iter_handles_test("Place", self.db.iter_place_handles)

    def test_iter_repository_handles(self):
        self.__iter_handles_test("Repository", self.db.iter_repository_handles)

    def test_iter_source_handles(self):
        self.__iter_handles_test("Source", self.db.iter_source_handles)

    def test_iter_citation_handles(self):
        self.__iter_handles_test("Citation", self.db.iter_citation_handles)

    def test_iter_media_handles(self):
        self.__iter_handles_test("Media", self.db.iter_media_handles)

    def test_iter_note_handles(self):
        self.__iter_handles_test("Note", self.db.iter_note_handles)

    def test_iter_tag_handles(self):
        self.__iter_handles_test("Tag", self.db.iter_tag_handles)

    ################################################################
    #
    # Test iter_* methods
    #
    ################################################################

    def __iter_objects_test(self, obj_class, iter_func):
        for obj in iter_func():
            self.assertIsInstance(obj, obj_class)

    def test_iter_people(self):
        self.__iter_objects_test(Person, self.db.iter_people)

    def test_iter_families(self):
        self.__iter_objects_test(Family, self.db.iter_families)

    def test_iter_events(self):
        self.__iter_objects_test(Event, self.db.iter_events)

    def test_iter_places(self):
        self.__iter_objects_test(Place, self.db.iter_places)

    def test_iter_repositories(self):
        self.__iter_objects_test(Repository, self.db.iter_repositories)

    def test_iter_sources(self):
        self.__iter_objects_test(Source, self.db.iter_sources)

    def test_iter_citations(self):
        self.__iter_objects_test(Citation, self.db.iter_citations)

    def test_iter_media(self):
        self.__iter_objects_test(Media, self.db.iter_media)

    def test_iter_notes(self):
        self.__iter_objects_test(Note, self.db.iter_notes)

    def test_iter_tags(self):
        self.__iter_objects_test(Tag, self.db.iter_tags)

    ################################################################
    #
    # Test default and initial person methods
    #
    ################################################################

    def test_no_default_handle(self):
        self.db.set_default_person_handle(None)
        handle = self.db.get_default_handle()
        self.assertIsNone(handle)
        person = self.db.get_default_person()
        self.assertIsNone(person)
        person = self.db.find_initial_person()
        self.assertIsInstance(person, Person)

    def test_default_handle(self):
        default_handle = self.handles["Person"][0]
        self.db.set_default_person_handle(default_handle)
        handle = self.db.get_default_handle()
        self.assertEqual(handle, default_handle)
        person = self.db.get_default_person()
        self.assertEqual(person.handle, default_handle)
        person = self.db.find_initial_person()
        self.assertEqual(person.handle, default_handle)

    ################################################################
    #
    # Test get_total method
    #
    ################################################################
    def test_get_total(self):
        total = sum([len(self.handles[obj_type]) for obj_type in self.handles.keys()])
        self.assertEqual(self.db.get_total(), total)


# -------------------------------------------------------------------------
#
# DbEmptyTest class
#
# -------------------------------------------------------------------------
class DbEmptyTest(unittest.TestCase):
    """
    Tests with an empty database.
    """

    @classmethod
    def setUpClass(cls):
        cls.db = make_database("sqlite")
        cls.db.load(":memory:")

    ################################################################
    #
    # Test metadata methods
    #
    ################################################################

    def test_metadata(self):
        self.db._set_metadata("test-key", "test-value")
        value = self.db._get_metadata("test-key")
        self.assertEqual(value, "test-value")

    def test_metadata_missing(self):
        value = self.db._get_metadata("missing-key")
        self.assertEqual(value, [])

    def test_metadata_default(self):
        value = self.db._get_metadata("missing-key", default="default-value")
        self.assertEqual(value, "default-value")

    ################################################################
    #
    # Test default and initial person methods
    #
    ################################################################

    def test_no_default_handle(self):
        handle = self.db.get_default_handle()
        self.assertIsNone(handle)
        person = self.db.get_default_person()
        self.assertIsNone(person)
        person = self.db.find_initial_person()
        self.assertIsNone(person)

    ################################################################
    #
    # Test researcher methods
    #
    ################################################################

    def test_researcher(self):
        res1 = Researcher()
        res1.street = "street"
        res1.locality = "locality"
        res1.city = "city"
        res1.county = "county"
        res1.state = "state"
        res1.country = "country"
        res1.postal = "postal"
        res1.phone = "phone"
        res1.name = "name"
        res1.addr = "addr"
        res1.email = "email"
        self.db.set_researcher(res1)
        res2 = self.db.get_researcher()
        self.assertEqual(res1.serialize(), res2.serialize())

    ################################################################
    #
    # Test name group mapping
    #
    ################################################################

    def test_name_group_mapping(self):
        self.db.set_name_group_mapping("Clark", "Clarke")

        self.assertTrue(self.db.has_name_group_key("Clark"))
        self.assertFalse(self.db.has_name_group_key("Baker"))

        for key in self.db.get_name_group_keys():
            self.assertTrue(self.db.has_name_group_key(key))

        mapping = self.db.get_name_group_mapping("Clark")
        self.assertEqual(mapping, "Clarke")

    ################################################################
    #
    # Test get_total method
    #
    ################################################################
    def test_get_total(self):
        self.assertEqual(self.db.get_total(), 0)


# -------------------------------------------------------------------------
#
# DbPersonTest class
#
# -------------------------------------------------------------------------
class DbPersonTest(unittest.TestCase):
    """
    Tests with some sample people.
    """

    @classmethod
    def setUpClass(cls):
        cls.db = make_database("sqlite")
        cls.db.load(":memory:")

    def __add_person(self, gender, first_name, surname, trans):
        person = Person()
        person.gender = gender
        name = person.primary_name
        name.first_name = first_name
        surname1 = Surname()
        surname1.surname = surname
        name.set_surname_list([surname1])
        self.all_surnames.append(surname)
        self.db.add_person(person, trans)

    def setUp(self):
        self.all_surnames = []
        with DbTxn("Add test objects", self.db) as trans:
            self.__add_person(Person.MALE, "John", "Allen", trans)
            self.__add_person(Person.MALE, "John", "Baker", trans)
            self.__add_person(Person.MALE, "John", "Clark", trans)
            self.__add_person(Person.FEMALE, "John", "Davis", trans)
            self.__add_person(Person.UNKNOWN, "John", "Evans", trans)
            self.__add_person(Person.FEMALE, "Mary", "Allen", trans)
            self.__add_person(Person.FEMALE, "Mary", "Baker", trans)
            self.__add_person(Person.FEMALE, "Mary", "Clark", trans)
            self.__add_person(Person.MALE, "Mary", "Davis", trans)
            self.__add_person(Person.OTHER, "Mary", "Evans", trans)

    def tearDown(self):
        with DbTxn("Remove test objects", self.db) as trans:
            for handle in self.db.get_person_handles():
                self.db.remove_person(handle, trans)

    ################################################################
    #
    # Test surname list
    #
    ################################################################

    def test_surname_list(self):
        surname_list = self.db.get_surname_list()
        for surname in surname_list:
            self.assertIn(surname, self.all_surnames)

    ################################################################
    #
    # Test gender stats
    #
    ################################################################

    def test_gender_stats(self):
        stats = self.db.genderStats
        self.assertEqual(stats.name_stats("John"), (3, 1, 1))
        self.assertEqual(stats.name_stats("Mary"), (1, 3, 1))
        self.db.save_gender_stats(stats)
        saved = self.db.get_gender_stats()
        self.assertEqual(saved["John"], (3, 1, 1))
        self.assertEqual(saved["Mary"], (1, 3, 1))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
# Copyright (C) 2014       Mathieu MD <mathieu.md@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#

"""
Japanese-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------

import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Japanese parser
#
# -------------------------------------------------------------------------
class DateParserJA(DateParser):
    """
    Convert a text string into a Date object. If the date cannot be
    converted, the text string is assigned.
    """

    # modifiers after the date
    modifier_after_to_int = {
        "以前": Date.MOD_BEFORE,
        "以降": Date.MOD_AFTER,
        "頃": Date.MOD_ABOUT,
        "ごろ": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }

    month_to_int = DateParser.month_to_int

    quality_to_int = {
        "およそ": Date.QUAL_ESTIMATED,
        "ごろ": Date.QUAL_ESTIMATED,
        "位": Date.QUAL_ESTIMATED,
        "の見積り": Date.QUAL_ESTIMATED,
        "計算上": Date.QUAL_CALCULATED,
    }

    bce = ["紀元前", "BC"] + DateParser.bce

    def init_strings(self):
        """
        This method compiles regular expression strings for matching dates.
        """
        DateParser.init_strings(self)

        DateParser.calendar_to_int.update(
            {
                "グレゴリオ暦": Date.CAL_GREGORIAN,
                "g": Date.CAL_GREGORIAN,
                "ユリウス暦": Date.CAL_JULIAN,
                "j": Date.CAL_JULIAN,
                "ユダヤ暦": Date.CAL_HEBREW,
                "h": Date.CAL_HEBREW,
                "ヒジュラ暦": Date.CAL_ISLAMIC,
                "i": Date.CAL_ISLAMIC,
                "フランス革命暦": Date.CAL_FRENCH,
                "共和暦": Date.CAL_FRENCH,
                "f": Date.CAL_FRENCH,
                "イラン暦": Date.CAL_PERSIAN,
                "p": Date.CAL_PERSIAN,
                "スウェーデン暦": Date.CAL_SWEDISH,
                "s": Date.CAL_SWEDISH,
            }
        )

        DateParser.month_to_int.update(
            {
                "一月": 1,
                "ichigatsu": 1,
                "睦月": 1,
                "mutsuki": 1,
                "二月": 2,
                "nigatsu": 2,
                "如月": 2,
                "kisaragi": 2,
                "衣更着": 2,
                "kinusaragi": 2,
                "三月": 3,
                "sangatsu": 3,
                "弥生": 3,
                "yayoi": 3,
                "四月": 4,
                "shigatsu": 4,
                "卯月": 4,
                "uzuki": 4,
                "五月": 5,
                "gogatsu": 5,
                "皐月": 5,
                "satsuki": 5,
                "早苗月": 5,
                "sanaetsuki": 5,
                "六月": 6,
                "rokugatsu": 6,
                "水無月": 6,
                "minazuki": 6,
                "七月": 7,
                "shichigatsu": 7,
                "文月": 7,
                "fumizuki": 7,
                "八月": 8,
                "hachigatsu": 8,
                "葉月": 8,
                "hazuki": 8,
                "九月": 9,
                "kugatsu": 9,
                "長月": 9,
                "nagatsuki": 9,
                "十月": 10,
                "jugatsu": 10,
                "jūgatsu": 10,
                "juugatsu": 10,
                "神無月": 10,
                "kannazuki": 10,
                "kaminazuki": 10,
                "神有月": 10,
                "神在月": 10,
                "kamiarizuki": 10,
                "十一月": 11,
                "juichigatsu": 11,
                "jūichigatsu": 11,
                "juuichigatsu": 11,
                "霜月": 11,
                "shimotsuki": 11,
                "十二月": 12,
                "junigatsu": 12,
                "jūnigatsu": 12,
                "juunigatsu": 12,
                "師走": 12,
                "shiwasu": 12,
            }
        )

        _span_1 = ["から", "~", "〜"]
        _span_2 = ["まで"]
        _range_1 = ["から", "と", "~", "〜"]
        _range_2 = ["までの間", "の間"]
        self._span = re.compile(
            r"(?P<start>.+)(%s)(?P<stop>.+)(%s)"
            % ("|".join(_span_1), "|".join(_span_2)),
            re.IGNORECASE,
        )
        self._range = re.compile(
            r"(?P<start>.+)(%s)(?P<stop>.+)(%s)"
            % ("|".join(_range_1), "|".join(_range_2)),
            re.IGNORECASE,
        )
        self._numeric = re.compile(r"((\d+)年\s*)?((\d+)月\s*)?(\d+)?日?\s*$")
        self._cal = re.compile(r"(.*?)\s*\(%s\)\s*(.*)" % self._cal_str, re.IGNORECASE)
        self._qual = re.compile(r"(.*?)\s*%s\s*(.*)" % self._qual_str, re.IGNORECASE)


# -------------------------------------------------------------------------
#
# Japanese display
#
# -------------------------------------------------------------------------
class DateDisplayJA(DateDisplay):
    """
    Japanese language date display class.
    """

    def formats_changed(self):
        """Allow overriding so a subclass can modify"""

        # Specify what is actually the "System Default".
        example = self.dhformat
        example = example.replace("%d", "31")
        example = example.replace("%m", "12")
        example = example.replace("%Y", "1999")

        # This definition must agree with its "_display_gregorian" method
        self.formats = (
            "YYYY-MM-DD (ISO)",  # 0
            "システムデフォールト (" + example + ")",  # 1
            "1999年12月31日",  # 2
            "1999年十二月31日",  # 3
        )

    def _display_gregorian(self, date_val, **kwargs):
        """
        display gregorian calendar date in different format
        """
        # this must agree with its locale-specific "formats" definition
        year = self._slash_year(date_val[2], date_val[3])
        if self.format == 0:
            # ISO
            return self.display_iso(date_val)

        elif self.format == 1:
            # System Default
            if date_val[2] < 0 or date_val[3]:
                return self.display_iso(date_val)
            else:
                if date_val[0] == date_val[1] == 0:
                    value = str(date_val[2])
                else:
                    value = self.dhformat.replace("%m", str(date_val[1]))
                    if date_val[0] == 0:  # ignore the zero day and its delimiter
                        i_day = value.find("%d")
                        value = value.replace(value[i_day : i_day + 3], "")
                    value = value.replace("%d", str(date_val[0]))
                    value = value.replace("%Y", str(date_val[2]))

        elif self.format == 2:
            # 1999年12月31日
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s年" % year
                else:
                    value = "%s年%s" % (year, self.short_months[date_val[1]])
            else:
                value = "%s年%s%s日" % (
                    year,
                    self.short_months[date_val[1]],
                    date_val[0],
                )

        elif self.format == 3:
            # 1999年十二月31日
            if date_val[0] == 0:
                if date_val[1] == 0:
                    value = "%s年" % year
                else:
                    value = "%s年%s" % (year, self.long_months[date_val[1]])
            else:
                value = "%s年%s%s日" % (
                    year,
                    self.long_months[date_val[1]],
                    date_val[0],
                )

        else:
            return self.display_iso(date_val)

        if date_val[2] < 0:
            return self._bce_str % value
        else:
            return value

    display = DateDisplay.display_formatted


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------

register_datehandler(
    ("ja_JP", "ja", "japanese", "Japanese", ("%Y年%m月%d日",)),
    DateParserJA,
    DateDisplayJA,
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# IsAncestorOf
#
# -------------------------------------------------------------------------
class IsAncestorOf(Rule):
    """Rule that checks for a person that is an ancestor of a specified person"""

    labels = [_("ID:"), _("Inclusive:")]
    name = _("Ancestors of <person>")
    category = _("Ancestral filters")
    description = _("Matches people that are ancestors of a specified person")

    def prepare(self, db: Database, user):
        """Assume that if 'Inclusive' not defined, assume inclusive"""
        self.db = db
        self.selected_handles: Set[str] = set()
        try:
            first = False if int(self.list[1]) else True
        except IndexError:
            first = True
        try:
            root_person = db.get_person_from_gramps_id(self.list[0])
            self.init_ancestor_list(db, root_person, first)
        except:
            pass

    def reset(self):
        self.selected_handles.clear()

    def apply_to_one(self, db: Database, person: Person) -> bool:
        return person.handle in self.selected_handles

    def init_ancestor_list(
        self, db: Database, person: Person | None, first: bool
    ) -> None:
        if not person:
            return
        if person.handle in self.selected_handles:
            return
        if not first:
            self.selected_handles.add(person.handle)
        fam_id = (
            person.parent_family_list[0] if len(person.parent_family_list) > 0 else None
        )
        if fam_id:
            fam = db.get_family_from_handle(fam_id)
            if fam:
                f_id = fam.father_handle
                m_id = fam.mother_handle

                if f_id:
                    self.init_ancestor_list(db, db.get_person_from_handle(f_id), False)
                if m_id:
                    self.init_ancestor_list(db, db.get_person_from_handle(m_id), False)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The TreeModel for the URL list in the Url Tab.
"""

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk


# -------------------------------------------------------------------------
#
# WebModel
#
# -------------------------------------------------------------------------
class WebModel(Gtk.ListStore):
    """
    WebModel derives from the ListStore, defining te items in the list
    """

    def __init__(self, obj_list, dbase):
        Gtk.ListStore.__init__(self, str, str, str, bool, object)
        self.db = dbase
        for obj in obj_list:
            self.append(row=[str(obj.type), obj.path, obj.desc, obj.get_privacy(), obj])



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import os

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.utils.file import media_path_full
from gramps.gen.datehandler import get_date
from gramps.gen.lib import Person
from gramps.gen.const import COLON, GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------

_YIELD_INTERVAL = 200

# ------------------------------------------------------------------------
#
# StatsGramplet class
#
# ------------------------------------------------------------------------


class StatsGramplet(Gramplet):
    """Display some statistics about the tree, with links to filtered quick
    reports
    """

    def init(self):
        self.set_text(_("No Family Tree loaded."))
        self.set_tooltip(_("Double-click item to see matches"))

    def db_changed(self):
        self.connect(self.dbstate.db, "person-add", self.update)
        self.connect(self.dbstate.db, "person-update", self.update)
        self.connect(self.dbstate.db, "person-delete", self.update)
        self.connect(self.dbstate.db, "family-add", self.update)
        self.connect(self.dbstate.db, "family-delete", self.update)
        self.connect(self.dbstate.db, "person-rebuild", self.update)
        self.connect(self.dbstate.db, "family-rebuild", self.update)

    def main(self):
        self.set_text(_("Processing..."))
        database = self.dbstate.db
        personList = database.iter_people()

        with_media = 0
        total_media = 0
        incomp_names = 0
        disconnected = 0
        missing_bday = 0
        males = 0
        females = 0
        others = 0
        unknowns = 0
        bytes_cnt = 0
        notfound = []

        mobjects = database.get_number_of_media()
        mbytes = "0"
        for media in database.iter_media():
            fullname = media_path_full(database, media.get_path())
            try:
                bytes_cnt += os.path.getsize(fullname)
                length = len(str(bytes_cnt))
                if bytes_cnt <= 999999:
                    mbytes = _("less than 1")
                else:
                    mbytes = str(bytes_cnt)[: (length - 6)]
            except OSError:
                notfound.append(media.get_path())

        for cnt, person in enumerate(personList):
            length = len(person.get_media_list())
            if length > 0:
                with_media += 1
                total_media += length

            for name in [person.get_primary_name()] + person.get_alternate_names():
                if name.get_first_name().strip() == "":
                    incomp_names += 1
                else:
                    if name.get_surname_list():
                        for surname in name.get_surname_list():
                            if surname.get_surname().strip() == "":
                                incomp_names += 1
                    else:
                        incomp_names += 1

            if (
                not person.get_main_parents_family_handle()
                and not person.get_family_handle_list()
            ):
                disconnected += 1

            birth_ref = person.get_birth_ref()
            if birth_ref:
                birth = database.get_event_from_handle(birth_ref.ref)
                if not get_date(birth):
                    missing_bday += 1
            else:
                missing_bday += 1

            if person.get_gender() == Person.FEMALE:
                females += 1
            elif person.get_gender() == Person.MALE:
                males += 1
            elif person.get_gender() == Person.OTHER:
                others += 1
            else:
                unknowns += 1
            if not cnt % _YIELD_INTERVAL:
                yield True
        self.clear_text()
        self.append_text(_("Individuals") + "\n")
        self.append_text("----------------------------\n")
        self.link(_("Number of individuals") + COLON, "Filter", "all people")
        self.append_text(" %s" % database.get_number_of_people())
        self.append_text("\n")
        self.link(_("%s:") % _("Males"), "Filter", "males")
        self.append_text(" %s" % males)
        self.append_text("\n")
        self.link(_("%s:") % _("Females"), "Filter", "females")
        self.append_text(" %s" % females)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Individuals with other gender"),
            "Filter",
            "people with other gender",
        )
        self.append_text(" %s" % others)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Individuals with unknown gender"),
            "Filter",
            "people with unknown gender",
        )
        self.append_text(" %s" % unknowns)
        self.append_text("\n")
        self.link(_("%s:") % _("Incomplete names"), "Filter", "incomplete names")
        self.append_text(" %s" % incomp_names)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Individuals missing birth dates"),
            "Filter",
            "people with missing birth dates",
        )
        self.append_text(" %s" % missing_bday)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Disconnected individuals"), "Filter", "disconnected people"
        )
        self.append_text(" %s" % disconnected)
        self.append_text("\n")
        self.append_text("\n%s\n" % _("Family Information"))
        self.append_text("----------------------------\n")
        self.link(_("%s:") % _("Number of families"), "Filter", "all families")
        self.append_text(" %s" % database.get_number_of_families())
        self.append_text("\n")
        if hasattr(database, "surname_list"):
            self.link(_("%s:") % _("Unique surnames"), "Filter", "unique surnames")
            self.append_text(" %s" % len(set(database.surname_list)))
            self.append_text("\n")
        self.append_text("\n%s\n" % _("Media Objects"))
        self.append_text("----------------------------\n")
        self.link(
            _("%s:") % _("Individuals with media objects"),
            "Filter",
            "people with media",
        )
        self.append_text(" %s" % with_media)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Total number of media object references"),
            "Filter",
            "media references",
        )
        self.append_text(" %s" % total_media)
        self.append_text("\n")
        self.link(
            _("%s:") % _("Number of unique media objects"), "Filter", "unique media"
        )
        self.append_text(" %s" % mobjects)
        self.append_text("\n")

        self.link(
            _("%s:") % _("Total size of media objects"), "Filter", "media by size"
        )
        self.append_text(" %s %s" % (mbytes, _("MB", "Megabyte")))
        self.append_text("\n")
        self.link(_("%s:") % _("Missing Media Objects"), "Filter", "missing media")
        self.append_text(" %s\n" % len(notfound))
        self.append_text("", scroll_to="begin")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import pickle

# -------------------------------------------------------------------------
#
# enable logging for error handling
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from ..ddtargets import DdTargets
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Pango
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.config import config
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.lib import ChildRef, Family, Name, NoteType, Person, Surname
from gramps.gen.db import DbTxn
from gramps.gen.errors import WindowActiveError
from gramps.gen.datehandler import displayer
from ..glade import Glade

from .editprimary import EditPrimary
from .editchildref import EditChildRef
from .editperson import EditPerson
from .displaytabs import (
    EmbeddedList,
    EventEmbedList,
    CitationEmbedList,
    FamilyAttrEmbedList,
    NoteTab,
    GalleryTab,
    FamilyLdsEmbedList,
    ChildModel,
    TEXT_COL,
    MARKUP_COL,
    ICON_COL,
)
from ..widgets import PrivacyButton, MonitoredEntry, MonitoredDataType, MonitoredTagList
from gramps.gen.plug import CATEGORY_QR_FAMILY
from ..dialog import ErrorDialog, RunDatabaseRepair, WarningDialog, MessageHideDialog
from gramps.gen.utils.db import (
    get_birth_or_fallback,
    get_death_or_fallback,
    get_marriage_or_fallback,
    preset_name,
    family_name,
)
from ..selectors import SelectorFactory
from gramps.gen.utils.id import create_id
from gramps.gen.const import URL_MANUAL_SECT1
from ..dbguielement import DbGUIElement

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1
WIKI_HELP_SEC = _("Family_Editor_dialog", "manual")

SelectPerson = SelectorFactory("Person")

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")
_LEFT_BUTTON = 1
_RIGHT_BUTTON = 3


class ChildEmbedList(DbGUIElement, EmbeddedList):
    """
    The child embed list is specific to the Edit Family dialog, so it
    is contained here instead of in displaytabs.
    """

    _HANDLE_COL = 15
    _DND_TYPE = DdTargets.CHILDREF
    _DND_EXTRA = DdTargets.PERSON_LINK

    _MSG = {
        "add": _("Create a new person and add the child to the family"),
        "del": _("Remove the child from the family"),
        "edit": _("Edit the child reference"),
        "share": _("Add an existing person as a child of the family"),
        "up": _("Move the child up in the children list"),
        "down": _("Move the child down in the children list"),
    }

    # (name, column in model, width, markup/text, font weight)
    _column_names = [
        (_("#"), 0, 25, TEXT_COL, -1, None),
        (_("ID"), 1, 60, TEXT_COL, -1, None),
        (_("Name"), 10, 250, TEXT_COL, -1, None),
        (_("Gender"), 3, 75, TEXT_COL, -1, None),
        (_("Paternal"), 4, 100, TEXT_COL, -1, None),
        (_("Maternal"), 5, 100, TEXT_COL, -1, None),
        (_("Birth Date"), 11, 150, MARKUP_COL, -1, None),
        (_("Death Date"), 12, 150, MARKUP_COL, -1, None),
        (_("Birth Place"), 8, 150, TEXT_COL, -1, None),
        (_("Death Place"), 9, 150, TEXT_COL, -1, None),
        None,
        None,
        None,
        (_("Source"), 13, 30, ICON_COL, -1, "gramps-source"),
        (_("Private"), 14, 30, ICON_COL, -1, "gramps-lock"),
    ]

    def __init__(self, dbstate, uistate, track, family, config_key):
        """
        Create the object, storing the passed family value
        """
        self.family = family
        DbGUIElement.__init__(self, dbstate.db)
        EmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("Chil_dren"),
            ChildModel,
            config_key,
            share_button=True,
            move_buttons=True,
        )

    def _connect_db_signals(self):
        """
        called on init of DbGUIElement, connect to db as required.
        """
        # note: event-rebuild closes the editors, so no need to connect to it
        self.callman.register_callbacks(
            {
                "person-update": self.person_change,  # change to person we track
                "person-delete": self.person_delete,  # delete of person we track
            }
        )
        self.callman.connect_all(keys=["person"])

    def person_change(self, *obj):
        """
        Callback method called when a tracked person changes (description
        changes...)
        """
        self.rebuild()

    def person_delete(self, hndls):
        """
        Callback method called when a tracked person is deleted.
        There are two possibilities:
        * a tracked non-workgroup person is deleted, just rebuilding the view
            will correct this.
        * a workgroup person is deleted. The person must be removed from the
            obj so that no inconsistent data is shown.
        """
        for handle in hndls:
            prefs = self.get_data()
            ref_list = [pref.ref for pref in prefs]
            indexlist = []
            last = -1
            while True:
                try:
                    last = ref_list.index(handle, last + 1)
                    indexlist.append(last)
                except ValueError:
                    break
            # remove the deleted workgroup persons from the object
            for index in reversed(indexlist):
                del prefs[index]
        # now rebuild the display tab
        self.rebuild()

    def get_popup_menu_items(self):
        return [
            (False, _("Edit child"), self.edit_child_button_clicked),
            (True, _("_Add"), self.add_button_clicked),
            (True, _("Add an existing child"), self.share_button_clicked),
            (False, _("Edit relationship"), self.edit_button_clicked),
            (True, _("_Remove"), self.del_button_clicked),
        ]

    def get_middle_click(self):
        return self.edit_child_button_clicked

    def get_icon_name(self):
        return "gramps-family"

    def get_data(self):
        """
        Normally, get_data returns a list. However, we return family
        object here instead.
        """
        prefs = self.family.get_child_ref_list()
        self.callman.register_handles({"person": [eref.ref for eref in prefs]})
        return prefs

    def column_order(self):
        return [
            (1, 13),
            (1, 14),
            (1, 0),
            (1, 1),
            (1, 2),
            (1, 3),
            (1, 4),
            (1, 5),
            (1, 6),
            (0, 8),
            (0, 9),
        ]

    def add_button_clicked(self, obj=None):
        person = Person()
        autoname = config.get("behavior.surname-guessing")
        # _("Father's surname"),
        # _("None"),
        # _("Combination of mother's and father's surname"),
        # _("Icelandic style"),
        if autoname == 0:
            name = self.north_american()
        elif autoname == 2:
            name = self.latin_american()
        else:
            name = self.no_name()
        person.set_primary_name(name)

        EditPerson(self.dbstate, self.uistate, self.track, person, self.new_child_added)

    def handle_extra_type(self, objtype, obj):
        """
        Called when a person is dropped onto the list.  objtype will be
        'person-link' and obj will contain a person handle.
        """
        person = self.dbstate.db.get_person_from_handle(obj)
        self.new_child_added(person)

    def new_child_added(self, person):
        ref = ChildRef()
        ref.ref = person.get_handle()
        self.family.add_child_ref(ref)
        self.rebuild()
        GLib.idle_add(
            self.tree.scroll_to_cell, len(self.family.get_child_ref_list()) - 1
        )
        self.call_edit_childref(ref)

    def child_ref_edited(self, person):
        self.rebuild()

    def share_button_clicked(self, obj=None):
        # it only makes sense to skip those who are already in the family
        skip_list = [self.family.get_father_handle(), self.family.get_mother_handle()]
        skip_list.extend(x.ref for x in self.family.get_child_ref_list())

        sel = SelectPerson(
            self.dbstate, self.uistate, self.track, _("Select Child"), skip=skip_list
        )
        person = sel.run()

        if person:
            ref = ChildRef()
            ref.ref = person.get_handle()
            self.family.add_child_ref(ref)
            self.rebuild()
            GLib.idle_add(
                self.tree.scroll_to_cell, len(self.family.get_child_ref_list()) - 1
            )
            self.call_edit_childref(ref)

    def run(self, skip):
        skip_list = [_f for _f in skip if _f]
        SelectPerson(
            self.dbstate, self.uistate, self.track, _("Select Child"), skip=skip_list
        )

    def del_button_clicked(self, obj=None):
        ref = self.get_selected()
        if ref:
            self.family.remove_child_ref(ref)
            self.rebuild()

    def edit_button_clicked(self, obj=None):
        ref = self.get_selected()
        if ref:
            self.call_edit_childref(ref)

    def call_edit_childref(self, ref):
        p = self.dbstate.db.get_person_from_handle(ref.ref)
        n = name_displayer.display(p)
        try:
            EditChildRef(
                n, self.dbstate, self.uistate, self.track, ref, self.child_ref_edited
            )
        except WindowActiveError:
            pass

    def edit_child_button_clicked(self, obj=None):
        ref = self.get_selected()
        if ref:
            p = self.dbstate.db.get_person_from_handle(ref.ref)
            try:
                EditPerson(
                    self.dbstate, self.uistate, self.track, p, self.child_ref_edited
                )
            except WindowActiveError:
                pass

    def north_american(self):
        """
        Child inherits name from father
        """
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        father_handle = self.family.get_father_handle()
        if father_handle:
            father = self.dbstate.db.get_person_from_handle(father_handle)
            preset_name(father, name)
        return name

    def no_name(self):
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        return name

    def latin_american(self):
        """
        Child inherits name from father and mother
        """
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        if self.family:
            father_handle = self.family.get_father_handle()
            father = (
                self.dbstate.db.get_person_from_handle(father_handle)
                if father_handle
                else None
            )
            mother_handle = self.family.get_mother_handle()
            mother = (
                self.dbstate.db.get_person_from_handle(mother_handle)
                if mother_handle
                else None
            )
            if not father and not mother:
                return name
            if not father:
                preset_name(mother, name)
                return name
            if not mother:
                preset_name(father, name)
                return name
            # we take first surname, and keep that
            mothername = Name()
            preset_name(mother, mothername)
            preset_name(father, name)
            mothersurname = mothername.get_surname_list()[0]
            mothersurname.set_primary(False)
            name.set_surname_list([name.get_surname_list()[0], mothersurname])
            return name
        else:
            return name


# -------------------------------------------------------------------------
#
# EditFamily
#
# -------------------------------------------------------------------------
class EditFamily(EditPrimary):
    QR_CATEGORY = CATEGORY_QR_FAMILY

    def __init__(self, dbstate, uistate, track, family, callback=None):
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            family,
            dbstate.db.get_family_from_handle,
            dbstate.db.get_family_from_gramps_id,
            callback,
        )

        # look for the scenerio of a child and no parents on a new
        # family

        if (
            self.added
            and not self.obj.get_father_handle()
            and not self.obj.get_mother_handle()
            and len(self.obj.get_child_ref_list()) == 1
        ):
            self.add_parent = True
            if not config.get("preferences.family-warn"):
                for i in self.hidden:
                    i.set_sensitive(False)

                MessageHideDialog(
                    _("Adding parents to a person"),
                    _(
                        "It is possible to accidentally create multiple "
                        "families with the same parents. To help avoid "
                        "this problem, only the buttons to select parents "
                        "are available when you create a new family. The "
                        "remaining fields will become available after you "
                        "attempt to select a parent."
                    ),
                    "preferences.family-warn",
                    parent=self.window,
                )
        else:
            self.add_parent = False

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        # FIXME, we rebind show_all below, this prevents garbage collection of
        #  the dialog, fix the rebind
        self.window.show_all = None
        EditPrimary._cleanup_on_exit(self)

    def empty_object(self):
        return Family()

    def _local_init(self):
        self.added = self.obj.handle is None
        if self.added:
            self.obj.handle = create_id()

        self.build_interface()
        self.load_data()

    def _connect_db_signals(self):
        """
        implement from base class DbGUIElement
        Register the callbacks we need.
        Note:
            * we do not connect to person-delete, as a delete of a person in
                the family outside of this editor will cause a family-update
                signal of this family
        """
        self.callman.register_handles({"family": [self.obj.get_handle()]})
        self.callman.register_callbacks(
            {
                "family-update": self.check_for_family_change,
                "family-delete": self.check_for_close,
                "family-rebuild": self._do_close,
                "event-update": self.topdata_updated,  # change eg birth event fath
                "event-rebuild": self.topdata_updated,
                "event-delete": self.topdata_updated,  # delete eg birth event fath
                "person-update": self.topdata_updated,  # change eg name of father
                "person-delete": self.person_delete,  # mother/father deleted?
                "person-rebuild": self._do_close,
            }
        )
        self.callman.connect_all(keys=["family", "event", "person"])

    def person_delete(self, handles):
        """This checks if mother/father is deleted, specifically when newly
        added before data is saved"""
        for hndl in handles:
            if self.obj.father_handle == hndl:
                self.obj.father_handle = None
            if self.obj.mother_handle == hndl:
                self.obj.mother_handle = None
            self.load_data()

    def check_for_family_change(self, handles):
        """
        Callback for family-update signal
        1. This method checks to see if the family shown has been changed. This
            is possible eg in the relationship view. If the family was changed,
            the view is refreshed and a warning dialog shown to indicate all
            changes have been lost.
            If a source/note/event is deleted, this method is called too. This
            is unfortunate as the displaytabs can track themself a delete and
            correct the view for this. Therefore, these tabs are not rebuild.
            Conclusion: this method updates so that remove/change of parent or
            remove/change of children in relationship view reloads the family
            from db.
        2. Changes in other families are of no consequence to the family shown
        """
        if self.obj.get_handle() in handles:
            # rebuild data
            ## Todo: Gallery and note tab are not rebuild ??
            objreal = self.dbstate.db.get_family_from_handle(self.obj.get_handle())
            # update selection of data that we obtain from database change:
            maindatachanged = (
                self.obj.gramps_id != objreal.gramps_id
                or self.obj.father_handle != objreal.father_handle
                or self.obj.mother_handle != objreal.mother_handle
                or self.obj.private != objreal.private
                or self.obj.type != objreal.type
                or self.obj.get_tag_list() != objreal.get_tag_list()
                or self.obj.child_ref_list != objreal.child_ref_list
            )
            if maindatachanged:
                self.obj.gramps_id = objreal.gramps_id
                self.obj.father_handle = objreal.father_handle
                self.obj.mother_handle = objreal.mother_handle
                self.obj.private = objreal.private
                self.obj.type = objreal.type
                self.obj.set_tag_list(objreal.get_tag_list())
                self.obj.child_ref_list = objreal.child_ref_list
                self.reload_people()

            # No matter why the family changed (eg delete of a source), we notify
            # the user
            WarningDialog(
                _("Family has changed"),
                _(
                    "The %(object)s you are editing has changed outside this editor."
                    " This can be due to a change in one of the main views, for "
                    "example a source used here is deleted in the source view.\n"
                    "To make sure the information shown is still correct, the "
                    "data shown has been updated. Some edits you have made may have"
                    " been lost."
                )
                % {"object": _("family")},
                parent=self.window,
            )

    def topdata_updated(self, *obj):
        """
        Callback method called if data shown in top part of family editor
        (a parent, birth/death event of parent) changes
        Note: person events shown in the event list are not tracked, the
              tabpage itself tracks it
        """
        self.load_data()

    def show_buttons(self):
        """
        Used to reshow hidden/showing buttons.
        """
        fhandle = self.obj.get_father_handle()
        self.update_father(fhandle)
        mhandle = self.obj.get_mother_handle()
        self.update_mother(mhandle)

    def reload_people(self):
        fhandle = self.obj.get_father_handle()
        self.update_father(fhandle)

        mhandle = self.obj.get_mother_handle()
        self.update_mother(mhandle)
        self.child_tab.rebuild()

    def get_menu_title(self):
        if self.obj and self.obj.get_handle():
            dialog_title = family_name(self.obj, self.db, _("New Family"))
            dialog_title = _("Family") + ": " + dialog_title
        else:
            dialog_title = _("New Family")
        return dialog_title

    def build_menu_names(self, family):
        return (_("Edit Family"), self.get_menu_title())

    def build_interface(self):
        self.top = Glade()
        self.set_window(self.top.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.family", 700, 500)

        # HACK: how to prevent hidden items from showing
        #       when you use show_all?
        # Consider using show() rather than show_all()?
        # FIXME: remove if we can use show()
        self.window.show_all = self.window.show

        self.fbirth = self.top.get_object("fbirth")
        self.fdeath = self.top.get_object("fdeath")
        self.fbirth_label = self.top.get_object("label578")
        self.fdeath_label = self.top.get_object("label579")

        self.mbirth = self.top.get_object("mbirth")
        self.mdeath = self.top.get_object("mdeath")
        self.mbirth_label = self.top.get_object("label567")
        self.mdeath_label = self.top.get_object("label568")

        self.mname = self.top.get_object("mname")
        self.fname = self.top.get_object("fname")

        self.mbutton_index = self.top.get_object("mbutton_index")
        self.mbutton_add = self.top.get_object("mbutton_add")
        self.mbutton_del = self.top.get_object("mbutton_del")
        self.mbutton_edit = self.top.get_object("mbutton_edit")

        self.mbutton_index.set_tooltip_text(_("Select a person as the mother"))
        self.mbutton_add.set_tooltip_text(_("Add a new person as the mother"))
        self.mbutton_del.set_tooltip_text(_("Remove the person as the mother"))

        self.mbutton_edit.connect("button-press-event", self.edit_mother)
        self.mbutton_edit.connect("key-press-event", self.edit_mother)
        self.mbutton_index.connect("clicked", self.sel_mother_clicked)
        self.mbutton_del.connect("clicked", self.del_mother_clicked)
        self.mbutton_add.connect("clicked", self.add_mother_clicked)

        self.fbutton_index = self.top.get_object("fbutton_index")
        self.fbutton_add = self.top.get_object("fbutton_add")
        self.fbutton_del = self.top.get_object("fbutton_del")
        self.fbutton_edit = self.top.get_object("fbutton_edit")

        self.fbutton_index.set_tooltip_text(_("Select a person as the father"))
        self.fbutton_add.set_tooltip_text(_("Add a new person as the father"))
        self.fbutton_del.set_tooltip_text(_("Remove the person as the father"))

        self.fbutton_edit.connect("button-press-event", self.edit_father)
        self.fbutton_edit.connect("key-press-event", self.edit_father)
        self.fbutton_index.connect("clicked", self.sel_father_clicked)
        self.fbutton_del.connect("clicked", self.del_father_clicked)
        self.fbutton_add.connect("clicked", self.add_father_clicked)

        if self.added:
            return  # avoids crash on drag because not in db yet
        # allow for a context menu
        self.set_contexteventbox(self.top.get_object("eventboxtop"))
        # allow for drag of the family object from eventboxtop
        self.contexteventbox.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [DdTargets.FAMILY_LINK.target()],
            Gdk.DragAction.COPY,
        )
        self.contexteventbox.drag_source_set_icon_name("gramps-family")
        self.contexteventbox.connect("drag_data_get", self.on_drag_data_get_family)

    def on_drag_data_get_family(self, widget, context, sel_data, info, time):
        if info == DdTargets.FAMILY_LINK.app_id:
            data = (DdTargets.FAMILY_LINK.drag_type, id(self), self.obj.get_handle(), 0)
            sel_data.set(DdTargets.FAMILY_LINK.atom_drag_type, 8, pickle.dumps(data))

    def _update_parent_dnd_handler(
        self, event_box, parent_handle, on_drag_data_get, on_drag_data_received
    ):
        """
        Set the drag action from the label of father when he exists
        """
        if parent_handle:
            # Allow drag
            if not event_box.drag_source_get_target_list():
                event_box.drag_source_set(
                    Gdk.ModifierType.BUTTON1_MASK,
                    [DdTargets.PERSON_LINK.target()],
                    Gdk.DragAction.COPY,
                )
                event_box.drag_source_set_icon_name("gramps-person")
                event_box.connect("drag_data_get", on_drag_data_get)
            # Disallow drop:
            if event_box.drag_dest_get_target_list():
                event_box.drag_dest_unset()
        else:
            # Disallow drag
            if event_box.drag_source_get_target_list():
                event_box.drag_source_unset()
            # allow for drop:
            if not event_box.drag_dest_get_target_list():
                event_box.drag_dest_set(
                    Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP,
                    [DdTargets.PERSON_LINK.target()],
                    Gdk.DragAction.COPY,
                )
                event_box.connect("drag_data_received", on_drag_data_received)

    def on_drag_fatherdata_get(self, widget, context, sel_data, info, time):
        if info == DdTargets.PERSON_LINK.app_id:
            data = (
                DdTargets.PERSON_LINK.drag_type,
                id(self),
                self.obj.get_father_handle(),
                0,
            )
            sel_data.set(DdTargets.PERSON_LINK.atom_drag_type, 8, pickle.dumps(data))

    def on_drag_motherdata_get(self, widget, context, sel_data, info, time):
        if info == DdTargets.PERSON_LINK.app_id:
            data = (
                DdTargets.PERSON_LINK.drag_type,
                id(self),
                self.obj.get_mother_handle(),
                0,
            )
            sel_data.set(DdTargets.PERSON_LINK.atom_drag_type, 8, pickle.dumps(data))

    def _connect_signals(self):
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_cancel_button(self.top.get_object("cancel"))
        # TODO help button (rename glade button name)
        self.define_help_button(
            self.top.get_object("button119"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _can_be_replaced(self):
        pass

    def _setup_fields(self):
        self.private = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.gid = MonitoredEntry(
            self.top.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.data_type = MonitoredDataType(
            self.top.get_object("marriage_type"),
            self.obj.set_relationship,
            self.obj.get_relationship,
            self.db.readonly,
            self.db.get_family_relation_types(),
        )

    def load_data(self):
        """
        Show top data of family editor: father and mother info
        and set self.phandles with all person handles in the family
        """
        fhandle = self.obj.get_father_handle()
        self.update_father(fhandle)

        mhandle = self.obj.get_mother_handle()
        self.update_mother(mhandle)

        self.phandles = [mhandle, fhandle]
        self.phandles.extend(x.ref for x in self.obj.get_child_ref_list())

        self.phandles = [_f for _f in self.phandles if _f]

    def get_start_date(self):
        """
        Get the start date for a family, usually a marriage date, or
        something close to marriage.
        """
        event = get_marriage_or_fallback(self.dbstate.db, self.obj)
        return event.get_date_object() if event else None

    def _create_tabbed_pages(self):
        notebook = Gtk.Notebook()

        self.child_list = ChildEmbedList(
            self.dbstate, self.uistate, self.track, self.obj, "family_editor_childrefs"
        )
        self.child_tab = self._add_tab(notebook, self.child_list)
        self.track_ref_for_deletion("child_list")
        self.track_ref_for_deletion("child_tab")

        self.event_list = EventEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj,
            "family_editor_events",
            start_date=self.get_start_date(),
        )

        self._add_tab(notebook, self.event_list)
        self.track_ref_for_deletion("event_list")

        self.citation_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "family_editor_citations",
            self.get_menu_title(),
        )
        self._add_tab(notebook, self.citation_list)
        self.track_ref_for_deletion("citation_list")

        self.attr_list = FamilyAttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "family_editor_attributes",
        )
        self._add_tab(notebook, self.attr_list)
        self.track_ref_for_deletion("attr_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "family_editor_notes",
            self.get_menu_title(),
            notetype=NoteType.FAMILY,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_media_list(),
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.lds_embed = FamilyLdsEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_lds_ord_list(),
            "family_editor_ldsord",
        )
        if not (config.get("interface.hide-lds") and self.lds_embed.is_empty()):
            self._add_tab(notebook, self.lds_embed)
        self.track_ref_for_deletion("lds_embed")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()

        self.hidden = (notebook, self.top.get_object("info"))
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def update_father(self, handle):
        self.load_parent(
            handle,
            self.fname,
            self.fbirth,
            self.fbirth_label,
            self.fdeath,
            self.fdeath_label,
            self.fbutton_index,
            self.fbutton_add,
            self.fbutton_del,
            self.fbutton_edit,
        )
        self._update_parent_dnd_handler(
            self.top.get_object("ftable_event_box"),
            self.obj.get_father_handle(),
            self.on_drag_fatherdata_get,
            self.on_drag_fatherdata_received,
        )

    def update_mother(self, handle):
        self.load_parent(
            handle,
            self.mname,
            self.mbirth,
            self.mbirth_label,
            self.mdeath,
            self.mdeath_label,
            self.mbutton_index,
            self.mbutton_add,
            self.mbutton_del,
            self.mbutton_edit,
        )
        self._update_parent_dnd_handler(
            self.top.get_object("mtable_event_box"),
            self.obj.get_mother_handle(),
            self.on_drag_motherdata_get,
            self.on_drag_motherdata_received,
        )

    def add_mother_clicked(self, obj):
        person = Person()
        person.set_gender(Person.FEMALE)
        autoname = config.get("behavior.surname-guessing")
        # _("Father's surname"),
        # _("None"),
        # _("Combination of mother's and father's surname"),
        # _("Icelandic style"),
        if autoname == 2:
            name = self.latin_american_child("mother")
        else:
            name = self.no_name()
        person.set_primary_name(name)
        EditPerson(
            self.dbstate, self.uistate, self.track, person, self.new_mother_added
        )

    def add_father_clicked(self, obj):
        person = Person()
        person.set_gender(Person.MALE)
        autoname = config.get("behavior.surname-guessing")
        # _("Father's surname"),
        # _("None"),
        # _("Combination of mother's and father's surname"),
        # _("Icelandic style"),
        if autoname == 0:
            name = self.north_american_child()
        elif autoname == 2:
            name = self.latin_american_child("father")
        else:
            name = self.no_name()
        person.set_primary_name(name)
        EditPerson(
            self.dbstate, self.uistate, self.track, person, self.new_father_added
        )

    def new_mother_added(self, person):
        for i in self.hidden:
            i.set_sensitive(True)
        self.obj.set_mother_handle(person.handle)
        self.update_mother(person.handle)

    def new_father_added(self, person):
        for i in self.hidden:
            i.set_sensitive(True)
        self.obj.set_father_handle(person.handle)
        self.update_father(person.handle)

    def del_mother_clicked(self, obj):
        for i in self.hidden:
            i.set_sensitive(True)

        self.obj.set_mother_handle(None)
        self.update_mother(None)

    def sel_mother_clicked(self, obj):
        for i in self.hidden:
            i.set_sensitive(True)

        sel = SelectPerson(
            self.dbstate,
            self.uistate,
            self.track,
            _("Select Mother"),
            search_or_filter=(4, _("Female")),
            skip=[x.ref for x in self.obj.get_child_ref_list()],
        )
        person = sel.run()

        if person:
            self.check_for_existing_family(
                self.obj.get_father_handle(), person.handle, self.obj.handle
            )
            self.obj.set_mother_handle(person.handle)
            self.update_mother(person.handle)

    def on_change_father(self, selector_window, obj):
        if obj.__class__ == Person:
            try:
                person = obj
                self.obj.set_father_handle(person.get_handle())
                self.update_father(person.get_handle())
            except:
                log.warning(
                    "Failed to update father: \n"
                    "obj returned from selector was: %s\n" % (repr(obj),)
                )
                raise

        else:
            log.warning(
                "Object selector returned obj.__class__ = %s, it should "
                "have been of type %s." % (obj.__class__.__name__, Person.__name__)
            )

        selector_window.close()

    def del_father_clicked(self, obj):
        for i in self.hidden:
            i.set_sensitive(True)

        self.obj.set_father_handle(None)
        self.update_father(None)

    def sel_father_clicked(self, obj):
        for i in self.hidden:
            i.set_sensitive(True)

        sel = SelectPerson(
            self.dbstate,
            self.uistate,
            self.track,
            _("Select Father"),
            search_or_filter=(4, _("Male")),
            skip=[x.ref for x in self.obj.get_child_ref_list()],
        )
        person = sel.run()
        if person:
            self.check_for_existing_family(
                person.handle, self.obj.get_mother_handle(), self.obj.handle
            )
            self.obj.set_father_handle(person.handle)
            self.update_father(person.handle)

    def check_for_existing_family(self, father_handle, mother_handle, family_handle):
        if father_handle:
            father = self.dbstate.db.get_person_from_handle(father_handle)
            ffam = set(father.get_family_handle_list())
            if mother_handle:
                mother = self.dbstate.db.get_person_from_handle(mother_handle)
                mfam = set(mother.get_family_handle_list())
                common = list(mfam.intersection(ffam))
                if len(common) > 0:
                    if self.add_parent or self.obj.handle not in common:
                        WarningDialog(
                            _("Duplicate Family"),
                            _(
                                "A family with these parents already exists "
                                "in the database. If you save, you will create "
                                "a duplicate family. It is recommended that "
                                "you cancel the editing of this window, and "
                                "select the existing family"
                            ),
                            parent=self.window,
                        )

    def edit_father(self, obj, event):
        handle = self.obj.get_father_handle()
        return self.edit_person(obj, event, handle)

    def edit_mother(self, obj, event):
        handle = self.obj.get_mother_handle()
        return self.edit_person(obj, event, handle)

    def edit_person(self, obj, event, handle):
        if button_activated(event, _LEFT_BUTTON):
            try:
                person = self.db.get_person_from_handle(handle)
                EditPerson(self.dbstate, self.uistate, self.track, person)
            except WindowActiveError:
                pass

    def load_parent(
        self,
        handle,
        name_obj,
        birth_obj,
        birth_label,
        death_obj,
        death_label,
        btn_index,
        btn_add,
        btn_del,
        btn_edit,
    ):
        # is a parent used here:
        is_used = handle is not None

        # now we display the area:
        if is_used:
            db = self.db
            person = db.get_person_from_handle(handle)
            name = "%s [%s]" % (name_displayer.display(person), person.gramps_id)
            birth = get_birth_or_fallback(db, person)
            self.callman.register_handles({"person": [handle]})
            if birth:
                # if event changes it view needs to update
                self.callman.register_handles({"event": [birth.get_handle()]})
                # Translators: needed for French, ignore otherwise
                birth_label.set_label(_("%s:") % birth.get_type())

            death = get_death_or_fallback(db, person)
            if death:
                # if event changes it view needs to update
                self.callman.register_handles({"event": [death.get_handle()]})
                # Translators: needed for French, ignore otherwise
                death_label.set_label(_("%s:") % death.get_type())

            btn_edit.set_tooltip_text(_("Edit %s") % name)
            btn_index.hide()
            btn_add.hide()
            btn_del.show()
            btn_edit.show()
        else:
            name = ""
            birth = None
            death = None

            btn_index.show()
            btn_add.show()
            btn_del.hide()
            btn_edit.hide()

        if name_obj:
            name_obj.set_text(name)
        if birth:
            birth_str = displayer.display(birth.get_date_object())
        else:
            birth_str = ""
        birth_obj.set_text(birth_str)
        if death:
            death_str = displayer.display(death.get_date_object())
        else:
            death_str = ""
        death_obj.set_text(death_str)

    def fix_parent_handles(self, orig_handle, new_handle, trans):
        if orig_handle != new_handle:
            if orig_handle:
                person = self.db.get_person_from_handle(orig_handle)
                person.remove_family_handle(self.obj.handle)
                self.db.commit_person(person, trans)
            if new_handle:
                person = self.db.get_person_from_handle(new_handle)
                person.add_family_handle(self.obj.handle)
                self.db.commit_person(person, trans)

    def on_drag_fatherdata_received(self, widget, context, x, y, sel_data, info, time):
        """
        Handle the standard gtk interface for drag_data_received.
        """
        if self.obj.get_father_handle():
            return
        for i in self.hidden:
            i.set_sensitive(True)
        if sel_data and sel_data.get_data():
            (drag_type, idval, handle, val) = pickle.loads(sel_data.get_data())
            person = self.db.get_person_from_handle(handle)

            if person:
                self.check_for_existing_family(
                    person.handle, self.obj.get_mother_handle(), self.obj.handle
                )
                self.obj.set_father_handle(person.handle)
                self.update_father(person.handle)

    def on_drag_motherdata_received(self, widget, context, x, y, sel_data, info, time):
        """
        Handle the standard gtk interface for drag_data_received.
        """
        if self.obj.get_mother_handle():
            return
        for i in self.hidden:
            i.set_sensitive(True)
        if sel_data and sel_data.get_data():
            (drag_type, idval, handle, val) = pickle.loads(sel_data.get_data())
            person = self.db.get_person_from_handle(handle)

            if person:
                self.check_for_existing_family(
                    self.obj.get_father_handle(), person.handle, self.obj.handle
                )
                self.obj.set_mother_handle(person.handle)
                self.update_mother(person.handle)

    def object_is_empty(self):
        return (
            not self.obj.get_father_handle()
            and not self.obj.get_mother_handle()
            and len(self.obj.get_child_ref_list()) == 0
        )

    def save(self, *obj):
        ## FIXME: how to catch a specific error?
        # try:
        self.__do_save()
        # except bsddb_db.DBRunRecoveryError as msg:
        #    RunDatabaseRepair(msg[1], parent=self.window)

    def __do_save(self):
        self.ok_button.set_sensitive(False)

        if not self.added:
            original = self.db.get_family_from_handle(self.obj.handle)
        else:
            original = None

        # do some basic checks

        child_list = [ref.ref for ref in self.obj.get_child_ref_list()]

        if self.obj.get_father_handle() in child_list:
            father = self.db.get_person_from_handle(self.obj.get_father_handle())
            name = "%s [%s]" % (name_displayer.display(father), father.gramps_id)
            ErrorDialog(
                _("A father cannot be his own child"),
                _("%s is listed as both the father and child " "of the family.") % name,
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return
        elif self.obj.get_mother_handle() in child_list:
            mother = self.db.get_person_from_handle(self.obj.get_mother_handle())
            name = "%s [%s]" % (name_displayer.display(mother), mother.gramps_id)
            ErrorDialog(
                _("A mother cannot be her own child"),
                _("%s is listed as both the mother and child " "of the family.") % name,
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        if not original and self.object_is_empty():
            ErrorDialog(
                _("Cannot save family"),
                _(
                    "No data exists for this family. "
                    "Please enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            msg1 = _("Cannot save family. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used. Please "
                "enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        # We disconnect the callbacks to all signals we connected earlier.
        # This prevents the signals originating in any of the following
        # commits from being caught by us again.
        self._cleanup_callbacks()

        if not original and not self.object_is_empty():
            with DbTxn(_("Add Family"), self.db) as trans:
                # find the father, add the family handle to the father
                handle = self.obj.get_father_handle()
                if handle:
                    parent = self.db.get_person_from_handle(handle)
                    parent.add_family_handle(self.obj.handle)
                    self.db.commit_person(parent, trans)

                # find the mother, add the family handle to the mother
                handle = self.obj.get_mother_handle()
                if handle:
                    parent = self.db.get_person_from_handle(handle)
                    parent.add_family_handle(self.obj.handle)
                    self.db.commit_person(parent, trans)

                # for each child, add the family handle to the child
                for ref in self.obj.get_child_ref_list():
                    child = self.db.get_person_from_handle(ref.ref)
                    # fix - relationships need to be extracted from the list
                    child.add_parent_family_handle(self.obj.handle)
                    self.db.commit_person(child, trans)

                self.db.add_family(self.obj, trans)
        elif self.data_has_changed():
            with DbTxn(_("Edit Family"), self.db) as trans:
                self.fix_parent_handles(
                    original.get_father_handle(), self.obj.get_father_handle(), trans
                )
                self.fix_parent_handles(
                    original.get_mother_handle(), self.obj.get_mother_handle(), trans
                )

                orig_set = set(original.get_child_ref_list())
                new_set = set(self.obj.get_child_ref_list())

                # remove the family from children which have been removed
                for ref in orig_set.difference(new_set):
                    person = self.db.get_person_from_handle(ref.ref)
                    person.remove_parent_family_handle(self.obj.handle)
                    self.db.commit_person(person, trans)

                # add the family to children which have been added
                for ref in new_set.difference(orig_set):
                    person = self.db.get_person_from_handle(ref.ref)
                    person.add_parent_family_handle(self.obj.handle)
                    self.db.commit_person(person, trans)

                if self.object_is_empty():
                    self.db.remove_family_relationships(self.obj.handle, trans)
                else:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_family_gramps_id())
                    self.db.commit_family(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj)
        self.callback = None

    def no_name(self):
        """
        Default surname guess.
        """
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        return name

    def north_american_child(self):
        """
        If SURNAME_GUESSING is north american, then find a child
        and return their name for the father.
        """
        # for each child, find one with a last name
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        for ref in self.obj.get_child_ref_list():
            child = self.db.get_person_from_handle(ref.ref)
            if child:
                preset_name(child, name)
                return name
        return name

    def latin_american_child(self, parent):
        """
        If SURNAME_GUESSING is latin american, then find a child
        and return their name for the father or mother.

        parent = "mother" | "father"
        """
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        # for each child, find one with a last name
        for ref in self.obj.get_child_ref_list():
            child = self.db.get_person_from_handle(ref.ref)
            if child:
                pname = child.get_primary_name()
                preset_name(child, name)  # add the known family surnames, etc.
                surnames = name.get_surname_list()
                if len(surnames) < 2:
                    return name
                else:
                    # return first for the father, and last for the mother
                    if parent == "father":
                        name.set_surname_list([surnames[0]])
                        return name
                    else:
                        name.set_surname_list([surnames[-1]])
                        return name
        return name


def button_activated(event, mouse_button):
    if (event.type == Gdk.EventType.BUTTON_PRESS and event.button == mouse_button) or (
        event.type == Gdk.EventType.KEY_PRESS and event.keyval in (_RETURN, _KP_ENTER)
    ):
        return True
    else:
        return False
